/*
 * File srvcore.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvcore.h
 *
 * server-side (and client-side) transport definitions and prototypes.
 *
 */
#ifndef SRVCORE_H_
#define SRVCORE_H_

#include "project.def"
#include "prolog.h"
#include "tinetype.h"
#include "errors.h"
#include "almlib.h"
#include "threader.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MULTITHREADED
#define gRunServerCycleInSeparateThread FALSE
#define gRunTransportInSeparateThread FALSE
#ifndef ReleaseSystemSemaphore
# define ReleaseSystemSemaphore(m) (m++)
#endif
#ifndef WaitForSystemSemaphore
# define WaitForSystemSemaphore(m,to) (m > 0 ? m-- : 0)
#endif
#ifndef ReleaseSystemMutex
# define ReleaseSystemMutex(m) (m = 0)
#endif
#ifndef WaitForSystemMutex
# define WaitForSystemMutex(m,to) ((m++) > 0 ? -1 : 0)
#endif
#else
extern int gRunServerCycleInSeparateThread;
extern int gRunTransportInSeparateThread;
extern int gNeedToStartSrvCycleThread;
extern int gNeedToStartClnCycleThread;
#endif
/* SetRunServerCycleInSeparateThread() */
/**
 * \fn TINE_EXPORT void SetRunServerCycleInSeparateThread(int value)
 * \brief Determines whether the server cycle is run in its own thread (boolean).
 *
 * If set to TRUE (default for multi-threaded builds), the server cycle will run in its
 * own thread, separate from the client cycle.  If set to FALSE, both cycles will run
 * in a round-robin fashion, where all socket descriptor sets are 'select()ed' on to assure
 * a prompt response.  This is usually fine.  However under extreme scenarios where performance
 * is of primary concern, then leaving this setting in its default TRUE state is the best
 * policy.  This of course only applies to multi-threaded builds of the TINE library.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  TRUE is
 * the default.
 *
 * \see GetRunServerCycleInSeparateThread().
 */
TINE_EXPORT void SetRunServerCycleInSeparateThread(int value);
/* GetRunServerCycleInSeparateThread() */
/**
 * \fn TINE_EXPORT int GetRunServerCycleInSeparateThread(void)
 * \brief Returns whether the server cycle is run in its own thread (boolean).
 *
 * If set to TRUE (default for multi-threaded builds), the server cycle will run in its
 * own thread, separate from the client cycle.  If set to FALSE, both cycles will run
 * in a round-robin fashion, where all socket descriptor sets are 'select()ed' on to assure
 * a prompt response.  This is usually fine.  However under extreme scenarios where performance
 * is of primary concern, then leaving this setting in its default TRUE state is the best
 * policy.  This of course only applies to multi-threaded builds of the TINE library.
 *
 * \return the current setting of this feature.
 *
 * \see SetRunServerCycleInSeparateThread().
 */
TINE_EXPORT int GetRunServerCycleInSeparateThread(void);
/* SetRunTransportInSeparateThread() */
/**
 * \fn TINE_EXPORT void SetRunTransportInSeparateThread(int value)
 * \brief Determines whether the (tcp) transport is run in its own thread (boolean).
 *
 * If set to TRUE (default for most multi-threaded builds), connected sockets will each have
 * their own thread where prompt data transport is assured.  If set to FALSE, all transport
 * delivery will follow the server contract access.  This (setting to FALSE) is generally an okay solution
 * and is not very resource intensive.  Setting this to TRUE will cause multiple buffering
 * of data payloads for each connected peer, which could cause memory problems on some OSes such as
 * VxWorks or other systems whether there is more concern about memory than speed.
 * The default setting is TRUE for windows, unix, and macos builds, and FALSE for VxWorks.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetRunServerCycleInSeparateThread().
 */
TINE_EXPORT void SetRunTransportInSeparateThread(int value);
/* GetRunTransportInSeparateThread() */
/**
 * \fn TINE_EXPORT int GetRunTransportInSeparateThread(void)
 * \brief Returns whether the (tcp) transport is run in its own thread (boolean).
 *
 * If set to TRUE (default for most multi-threaded builds), connected sockets will each have
 * their own thread where prompt data transport is assured.  If set to FALSE, all transport
 * delivery will follow the server contract access.  This (setting to FALSE) is generally an okay solution
 * and is not very resource intensive.  Setting this to TRUE will cause multiple buffering
 * of data payloads for each connected peer, which could cause memory problems on some OSes such as
 * VxWorks or other systems whether there is more concern about memory than speed.
 * The default setting is TRUE for windows, unix, and macos builds, and FALSE for VxWorks.
 *
 * \return the current setting of this feature.
 *
 * \see SetRunTransportInSeparateThread().
 */
TINE_EXPORT int GetRunTransportInSeparateThread(void);

extern int gAllowBkgTaskReentrancy;
/* SetAllowBackgroundTaskReentrancy() */
/**
 * \fn TINE_EXPORT void SetAllowBackgroundTaskReentrancy(int value)
 * \brief Determines whether equipment module background tasks may be re-entered (boolean).
 *
 * If set to TRUE, the server's background task will allow re-entrancy
 * which might occur if the task rate instructs the background task to be called
 * even though the previous call has not yet returned.  In such cases the programmer
 * must deal with whatever reentrancy issues there might be.  The default is value is FALSE.
 * (not reentrant).
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetAllowBackgroundTaskReentrancy().
 */
TINE_EXPORT void SetAllowBackgroundTaskReentrancy(int value);

/* GetAllowBackgroundTaskReentrancy() */
/**
 * \fn TINE_EXPORT int GetAllowBackgroundTaskReentrancy(void)
 * \brief Returns whether equipment module background tasks may be re-entered (boolean).
 *
 * If set to TRUE, the server's background task will allow re-entrancy
 * which might occur if the task rate instructs the background task to be called
 * even though the previous call has not yet returned.  In such cases the programmer
 * must deal with whatever reentrancy issues there might be.  This routine returns the
 * current setting.
 *
 * \return the current setting of this feature.
 *
 * \see SetAllowBackgroundTaskReentrancy().
 */
TINE_EXPORT int GetAllowBackgroundTaskReentrancy(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedBackgroundTasks(int value)
 * \brief Determines whether equipment module background tasks are to run in separate threads (boolean).
 *
 * If set to TRUE, the server's background task will run in its own thread, where a multi-threaded
 * build of the library is in use.  Note that this could result in data concurrency issues if the
 * background task is acquiring data which is to be used by an equipment module's equipment function.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.
 * TRUE is the default.
 *
 * \see SetAllowBackgroundTaskReentrancy(), GetUseMultiThreadedBackgroundTasks().
 */
TINE_EXPORT void SetUseMultiThreadedBackgroundTasks(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedBackgroundTasks(void)
 * \brief Returns whether equipment module background tasks are to run in separate threads (boolean).
 *
 * If set to TRUE, the server's background task will run in its own thread, where a multi-threaded
 * build of the library is in use.  Note that this could result in data concurrency issues if the
 * background task is acquiring data which is to be used by an equipment module's equipment function.
 *
 * \return the current setting of this feature.
 *
 * \see SetAllowBackgroundTaskReentrancy(), SetUseMultiThreadedBackgroundTasks().
 */
TINE_EXPORT int GetUseMultiThreadedBackgroundTasks(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedEquipmentFunctions(int value)
 * \brief Determines whether an equipment module equipment function can run in a separate threads (boolean).
 *
 * If set to TRUE, the an equipment function will handle designated properties in a separate thread,
 * where a multi-threaded build of the library is in use.
 * Ordinarily an equipment function should be regarded as an interrupt service routine which should return
 * 'promptly' after being called.  All properties will be directed to the equipment function using the same
 * thread as the server cycle, unless otherwise specified.  On the other hand if a property is specifically
 * registered to run the equipment module in a separate thread (because the action is known to require
 * significant time to complete) then it will run in its own thread if this setting has been enabled.
 * This will result in other calls to the equipment function receiving an 'operation busy' completion
 * while the equipment function is handling the time-consuming call.  Otherwise, concurrent calls will
 * simply time out at the client side.  To configure a property to be called in a separate thread one can
 * either call SetCallPropertyInSeparateThread() following the property registration of supply an "MT" switch
 * in the properties 'access' string which registering via file (either exports.csv or fec.xml).
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.
 * FALSE is the default.
 *
 * \see GetUseMultiThreadedEquipmentFunctions(), SetUseMultiThreadedStockFunctions(), SetCallPropertyInSeparateThread().
 */
TINE_EXPORT void SetUseMultiThreadedEquipmentFunctions(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedEquipmentFunctions(void)
 * \brief Returns whether an equipment module equipment function can run in a separate threads (boolean).
 *
 * If set to TRUE, the an equipment function will handle designated properties in a separate thread,
 * where a multi-threaded build of the library is in use.
 * Ordinarily an equipment function should be regarded as an interrupt service routine which should return
 * 'promptly' after being called.  All properties will be directed to the equipment function using the same
 * thread as the server cycle, unless otherwise specified.  On the other hand if a property is specifically
 * registered to run the equipment module in a separate thread (because the action is known to require
 * significant time to complete) then it will run in its own thread if this setting has been enabled.
 * This will result in other calls to the equipment function receiving an 'operation busy' completion
 * while the equipment function is handling the time-consuming call.  Otherwise, concurrent calls will
 * simply time out at the client side.  To configure a property to be called in a separate thread one can
 * either call SetCallPropertyInSeparateThread() following the property registration of supply an "MT" switch
 * in the properties 'access' string which registering via file (either exports.csv or fec.xml).
 *
 * \return the current setting of this feature.
 *
 * \see SetUseMultiThreadedEquipmentFunctions(), SetUseMultiThreadedStockFunctions(), SetCallPropertyInSeparateThread().
 */
TINE_EXPORT int GetUseMultiThreadedEquipmentFunctions(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedStockFunctions(int value)
 * \brief Determines whether stock propery calls can run in a separate threads (boolean).
 *
 * If set to TRUE, certain stock property calls will be handled in a separate thread,
 * where a multi-threaded build of the library is in use.
 * Ordinarily stock property calls return 'promptly' after being called, and make use of the same
 * thread as the server cycle, unless otherwise specified.  On the other hand certain stock properties
 * which make heavy use of the local disk might require considerable time to complete, noteably calls to
 * the local history subsystem or file retrieval calls.
 * A calls to such property will run in its own thread if this setting has been enabled.
 * This will result in other calls to the same stock property receiving an 'operation busy' completion
 * while the system is handling the time-consuming call.  Otherwise, concurrent calls will
 * simply time out at the client side.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.
 * TRUE is the default.
 *
 * \see GetUseMultiThreadedStockFunctions().
 */
TINE_EXPORT void SetUseMultiThreadedStockFunctions(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedStockFunctions(void)
 * \brief Returns whether stock propery calls can run in a separate threads (boolean).
 *
 * If set to TRUE, certain stock property calls will be handled in a separate thread,
 * where a multi-threaded build of the library is in use.
 * Ordinarily stock property calls return 'promptly' after being called, and make use of the same
 * thread as the server cycle, unless otherwise specified.  On the other hand certain stock properties
 * which make heavy use of the local disk might require considerable time to complete, noteably calls to
 * the local history subsystem or file retrieval calls.
 * A calls to such property will run in its own thread if this setting has been enabled.
 * This will result in other calls to the same stock property receiving an 'operation busy' completion
 * while the system is handling the time-consuming call.  Otherwise, concurrent calls will
 * simply time out at the client side.
 *
 * \return the current setting of this feature.
 *
 * \see SetUseMultiThreadedStockFunctions().
 */
TINE_EXPORT int GetUseMultiThreadedStockFunctions(void);

/**
 * \fn TINE_EXPORT int SetServerThreadPriority(int priority)
 * \brief Determines the priority of the server cycle thread as well as other associated server-side threads.
 *
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).
 *
 * \return the assigned server thread priority.
 *
 * \see GetServerThreadPriority().
 */
TINE_EXPORT int SetServerThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetServerThreadPriority(void)
 * \brief Returns the priority of the server cycle thread as well as other associated server-side threads.
 *
 * \return the server thread priority.
 *
 * \see SetServerThreadPriority().
 */
TINE_EXPORT int GetServerThreadPriority(void);

/**
 * \fn TINE_EXPORT int SetClientThreadPriority(int priority)
 * \brief Determines the priority of the client cycle thread as well as other associated client-side threads.
 *
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).
 *
 * \return the assigned client thread priority.
 *
 * \see GetClientThreadPriority().
 */
TINE_EXPORT int SetClientThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetClientThreadPriority(void)
 * \brief Returns the priority of the client cycle thread as well as other associated client-side threads.
 *
 * \return the client thread priority.
 *
 * \see SetClientThreadPriority().
 */
TINE_EXPORT int GetClientThreadPriority(void);

/**
 * \fn TINE_EXPORT int SetBackgroundThreadPriority(int priority)
 * \brief Determines the priority of any registered background threads.
 *
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).
 *
 * \return the assigned background thread priority.
 *
 * \see GetBackgroundThreadPriority().
 */
TINE_EXPORT int SetBackgroundThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetBackgroundThreadPriority(void)
 * \brief Returns the priority of the registered background threads
 *
 * \return the background thread priority.
 *
 * \see SetBackgroundThreadPriority().
 */
TINE_EXPORT int GetBackgroundThreadPriority(void);

/**
 * \fn TINE_EXPORT void SetStreamTransportRetryLimit(int value)
 * \brief Sets the stream transport retry limit
 *
 * When a server attempts to deliver data via a stream (TCP) connection
 * to a client it attempts to find available TCP send buffers to deliver the
 * data.  This generally is successful upon the initial attempt but can fail
 * due to network or buffering congestion.  In such cases a delay of a clock
 * tick and a retry will then succeed.  The number of times these retries are
 * allowed to happen are influenced by this parameter setting.  The default
 * value of '2' is generally correct, but under some conditions (one sluggish
 * client) can impede transfer to other clients and in such cases it might be
 * advisible to set this value to a lower value or even to '0' so as not to
 * make all attached clients suffer.  This value setting can be established
 * by making  use of this call.
 *
 * \see GetStreamTransportRetryLimit().
 */
TINE_EXPORT void SetStreamTransportRetryLimit(int value);

/**
 * \fn TINE_EXPORT int GetStreamTransportRetryLimit(void)
 * \brief Returns the stream transport retry limit
 *
 * When a server attempts to deliver data via a stream (TCP) connection
 * to a client it attempts to find available TCP send buffers to deliver the
 * data.  This generally is successful upon the initial attempt but can fail
 * due to network or buffering congestion.  In such cases a delay of a clock
 * tick and a retry will then succeed.  The number of times these retries are
 * allowed to happen are influenced by this parameter setting.  The default
 * value of '2' is generally correct, but under some conditions (one sluggish
 * client) can impede transfer to other clients and in such cases it might be
 * advisible to set this value to a lower value or even to '0' so as not to
 * make all attached clients suffer.
 *
 * \return the current setting for the stream transport retry limit (default = 2).
 *
 * \see SetStreamTransportRetryLimit().
 */
TINE_EXPORT int GetStreamTransportRetryLimit(void);

/**
 * \fn TINE_EXPORT int SetKernelPriority(int priority)
 * \brief Determines the priority of TINE kernel threads.
 *
 * Use this routine to set the priority of all TINE kernel threads (server-side, client-side,
 * background tasks) to the desired value.  This is primarily a convenience routine and is
 * equivalent to calling SetServerThreadPriority(), SetClientThreadPriority() and
 * SetBackgroundThreadPriority() individually.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).
 *
 * \return the assigned thread priority.
 *
 * \see SetServerThreadPriority(), SetClientThreadPriority(), SetBackgroundThreadPriority()
 */
TINE_EXPORT int SetKernelPriority(int priority);

int WaitForSyncLink(int id,int to);
int InitSystemMutexSet(void);

/* lockEquipmentModules() */
/**
 * \fn TINE_EXPORT int LockEquipmentModules(void)
 * \brief Locks all equipment modules
 *
 * Applying this call seizes a system mutex which will not allow the scheduler to cycle
 * thru the equipment module dispatch routines.
 * Consider setting the IDLE state of the module in lieu
 * of locking and unlocking.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see UnlockEquipmentModules().
 */
TINE_EXPORT int LockEquipmentModules(void);

/* UnlockEquipmentModules() */
/**
 * \fn TINE_EXPORT int UnlockEquipmentModules(void)
 * \brief Unlocks all equipment modules
 *
 * If the equipment modules were locked by applying a call to LockEquipmentModules()
 * then this call frees the lock state, allowing equipment module dispatch routines and
 * background tasks to continue.  Consider setting the IDLE state of the module in lieu
 * of locking and unlocking.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see LockEquipmentModules().
 */
TINE_EXPORT int UnlockEquipmentModules(void);

extern int gTransportCeiling;

/* SetServerTransportCeiling() */
/**
 * \fn TINE_EXPORT void SetServerTransportCeiling(int value)
 *
 * \brief Sets the server-side transport ceiling (in bytes) which gives the maximum transport size for
 * supported by calls to a multi-threaded server.
 *
 * Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
 * of the caller's requested data size to be used in data comparisons, the other buffer being in
 * the case of multi-threaded servers,
 * another dynamic allocation of the data size to be used in dispatch calls.
 * In order to allow transfer of 'large' data sets, this size
 * size needs to be adjusted accordingly. If not set, the setting given by the server work area
 * size will be used.
 *
 * \note This setting essentially establishes a limit on the allowed
 * output data size of client requests to multi-threaded servers, and prevents errant client
 * calls from monopolizing server memory.
 *
 * \note Stock properties will continue to make use of the server work area size as the allowed
 * data size.
 *
 * \param value is the size in bytes to be used as the transport ceiling for
 * any property supported by the server.
 * (default: the same value as the server work area size, i.e. 64 KBytes on most operating systems).
 *
 * \see GetServerTransportCeiling(), SetWorkAreaSize()
 */
TINE_EXPORT void SetServerTransportCeiling(int value);

/* GetServerTransportCeiling() */
/**
 * \fn TINE_EXPORT int GetServerTransportCeiling(void)
 *
 * \brief Gets the server-side transport ceiling (in bytes) which gives the maximum transport size for
 * supported by calls to a multi-threaded server.
 *
 * Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
 * of the caller's requested data size to be used in data comparisons, the other buffer being in
 * the case of multi-threaded servers,
 * another dynamic allocation of the data size to be used in dispatch calls.
 * In order to allow transfer of 'large' data sets, this size
 * size needs to be adjusted accordingly. If not set, the setting given by the server work area
 * size will be used. This call returns the current setting of this value.  If the call returns
 * '-1' this indicates that there is no setting and in such cases the server work area size is
 * used as the allowed output data size in bytes.
 *
 * \note This setting essentially establishes a limit on the allowed
 * output data size of client requests to multi-threaded servers, and prevents errant client
 * calls from monopolizing server memory.
 *
 * \return The size in bytes to be used as the transport ceiling for
 *  any property supported by the server. If the call returns
 * '-1' this indicates that there is no setting and in such cases the server work area size is
 * used as the allowed output data size in bytes.
 *
 * \see SetServerTransportCeiling(), GetWorkAreaSize()
 */
TINE_EXPORT int GetServerTransportCeiling(void);

#define LEGACY_TRANSPORT_PROTOCOL_LEVEL 6
#define DEFAULT_TRANSPORT_PROTOCOL_LEVEL 7

/**
 * \internal
 * \brief Default value for MaxNumConnections.
 */
#ifndef CONTBL_CAPACITY
# define CONTBL_CAPACITY 1024 /* connection table entries  */
#endif
extern int gAllowAutoGrowConnectionTable;

TINE_EXPORT void SetAllowAutoGrowConnectionTable(int value);

/**
 * \internal
 * \brief Communication protocol macro definitions ...
 */
#define IPX             0x01
#define SPX             0x02
#define TCP             0x04
#define UDP             0x08
#define PIPE            0x10
#define MMF             0x20
#define STREAM          0x40
#define DBGPIPE         0x80
#define TCP6            0x100
#define UDP6            0x200
#define STREAM6         0x400
#define DBGPIPE6        0x800

#define getProtocolAsString(i) (\
  (i) == IPX ? "IPX" :\
  (i) == SPX ? "SPX" :\
  (i) == TCP ? "TCP" :\
  (i) == UDP ? "UDP" :\
  (i) == PIPE ? "LCL" :\
  (i) == MMF ? "LCL" :\
  (i) == STREAM ? "STREAM" :\
  (i) == DBGPIPE ? "DBGPIPE" :\
  (i) == TCP6 ? "TCP6" :\
  (i) == UDP6 ? "UDP6" :\
  (i) == STREAM6 ? "STREAM6" :\
  (i) == DBGPIPE6 ? "DBGPIPE6" :\
  "---"\
)

#define CLN_REQ_ 0x01
#define SRV_RSP_ 0x02

#define GENS_TAG "GRPEQM"
#define ENS_TAG "ENSEQM"
#define CAS_TAG "CASEQM"
#define INTEQM_TAG "INTEQM"
#define GENS_FEC "GENS"
#define ENS_FEC "ENS"

#define USERS_LABEL "USERS"

#define FAILOVER_MASTER_DESC "failover MASTER"
#define FAILOVER_SLAVE_DESC "failover SLAVE"

#define FAILOVER_NONE   0x00
#define FAILOVER_MASTER 0x01
#define FAILOVER_SLAVE  0x02

#define FAILOVER_THRESHOLD 5
#define FAILOVER_POLLING_RATE 1000

#define CAS_SRV "CAS"
#define TIME_SRV "TIMESRV"
#define ENS_SRV  "ENS"
#define GENS_SRV "GENS"

#ifndef DOOCSEQM_TAG
# define DOOCSEQM_TAG "DCSEQM"
#endif

#define FECMF_FILENAME "fecmf.csv"

/**
 * \internal
 * \brief Some macro definitions for the UNIX world
 *
 */
#ifdef UNIX
# define S_DENYNONE (S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IWOTH|S_IROTH|S_IXOTH)
# define S_USERWRITE (S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)
# define S_ALLWRITE (S_IWUSR|S_IRUSR|S_IWGRP|S_IRGRP|S_IWOTH|S_IROTH)
#endif
/**
 * \internal
 * \brief The default value for MaxPollingRate.
 *
 * Client contracts asking to poll properties at an interval greater than MaxPollingRate will
 * be automatically converted to MaxPollingRate.
 * This in effect sets the upper limit for the timeout parameter for select() when SELECTING
 * mode is in effect.  Note that 'Rate' is actually a misnomer as the value of this parameter
 * actually gives an interval and not a rate.
 *
 */
#ifndef DEFAULT_POLLRATE
#  ifndef SYSPOLL
#    define DEFAULT_POLLRATE 10
#  else
#    define DEFAULT_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default value for MinPollingRate.
 *
 * Client contracts asking to poll properties at an interval smaller than MinPollingRate will
 * be automatically converted to MinPollingRate.  Thus a renegade client cannot force a server
 * to poll a property at smaller intervals.
 *
 */
#ifdef WIN32
# define MIN_POLLRATE 10
#endif
#ifndef MIN_POLLRATE
#  ifndef SYSPOLL
#    define MIN_POLLRATE 20
#  else
#    define MIN_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief enact nanosleep() for those OSes which don't have it
 *
 */
#if defined(VMS)|defined(WIN32)
#ifdef NEED_TIMESPEC
struct timespec
{
   time_t tv_sec; /* seconds */
   long   tv_nsec; /* nanoseconds */
};
#endif
#ifdef NEED_NANOSLEEP
TINE_EXPORT int nanosleep(const struct timespec *req, struct timespec *rem);
#endif
#endif

/**
 * \internal
 * \brief The default timeout used when ExecLink() is called.
 *
 * ExecLink() calls ExecLinkEx() with DEFAULT_CLIENT_POLLRATE as the timeout parameter.
 *
 */
#define DEFAULT_CLIENT_POLLRATE 1000

/**
 * \internal
 * \brief The default value for MaxSystemTransportSize
 *
 * Buffer space is reserved at initialization according to this value (64 KBytes).
 * Servers which need to support larger data transfers should set this value
 * accordingly.
 *
 * \todo Make this gracefully dynamic if possible. (Watch out for VxWorks, NIOS).
 *
 */
#define MAX_TRANSPORTSIZE 0xffff

 /**
 * \internal
 * \brief The default value of gBurstLimit.
 *
 * This corresponds to the number of data packets sent without a break.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gBurstLimit to a
 * high value on initialization.
 *
 */
#ifndef BURST_LIMIT
# define BURST_LIMIT 1000
#endif
extern int gBurstLimit;
extern int gTcpBurstLimit;

 /**
 * \internal
 * \brief The default value of gCycleDelay.
 *
 * This corresponds to the number of milliseonds to delay following a burst transmission.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gCycleDelay to a
 * smaller value (or 0) on initialization.
 *
 */
#ifndef CYCLE_DELAY
# define CYCLE_DELAY   1
#endif
extern int gCycleDelay;

extern int gCycleMicroDelay;

/**
 * \internal
 * \brief The default value of gSystemTick.
 *
 * Used internally within the TINE engine scheduler.
 *
 */
#ifndef SYSTEM_TICK
# define SYSTEM_TICK 10
#endif
extern UINT32 gSystemTick;

/**
 * \internal
 * \brief The default value of gSynchronizeContracts.
 *
 * Determines whether the TINE engine scheduler should attempt to
 * force contracts with polling intervals of (modulo) 1000 msec
 * should be called consecutively within the same cycle.  This
 * greatly improves the 'packing' of contracts being returned to
 * a given client.
 *
 */
#ifndef CONTRACT_SYNCHRONIZE
# define CONTRACT_SYNCHRONIZE TRUE
#endif
extern UINT32 gSynchronizeContracts;
extern int gSystemExitCode;
extern int gSystemRunningStandAlone;
extern int gSchedulePropertyLazy;
extern int gClientBurstLimit;
extern int gAutoLinkWatchdogs;
extern int allowCommonLinks;
extern int gAlwaysRetry;
extern int gServerCycleIdle;
extern int gEndianness;
extern int gCharEncoding;
extern int gSystemCycleIdleTime;
extern int EquipmentModulesInitialized;

/**
 * \internal
 * \brief Used as a flag in lieu of the startup time field in a contract.
 *
 * A server servicing a NETWORK subscription will set the startup time
 * field in the contract header to BCAST_ID (for either a multicast or
 * broadcast request) as a special signature.
 *
 */
#define BCAST_ID 255

/**
 * \internal
 * \brief Special scheduling signals for the TINE scheduler.
 *
 */
#define PRP_SCHEDULE_SIGNAL 0
#define PRP_REQUEST_SIGNAL 1
#define PRP_CANCEL_SIGNAL 2

/**
 * \internal
 * \brief Special reserved system names ...
 *
 */
#define SRVEQM_NAME "_SRV__"
#define SRVFEC_NAME "NETWORK"
#define SRVEXP_NAME "NETWORK"
#define STKEQM_NAME "_STK__"
/**
 * \internal
 * \brief Determine whether this request is coming from a NETWORK request
 * for services (e.g. ENS, CAS, etc.).
 *
 */
#define isServiceRequest(pctr) (strcmp((pctr)->EqmName,SRVEQM_NAME) == 0 ? TRUE : FALSE)

/**
 * \internal
 * \brief The default value for gRespondToServiceRequests.
 *
 * Determine whether this server can respond to Service requests.
 * Typically only the ENS will have this features turned on.  (Or possibly the CAS)
 *
 */
#ifndef RESPOND_TO_SERVICE_REQUESTS
# define RESPOND_TO_SERVICE_REQUESTS FALSE
#endif
extern TINE_EXPORT int gRespondToServiceRequests;

#define REGISTERED_ENS    0x01
#define REGISTERED_GENS   0x02
#define REGISTERED_MASTER 0x04
#define REGISTERED_ALL    0x07

#define TYPE_PRP 0x01
#define TYPE_DEV 0x02

TINE_EXPORT void SetForwardWildcardCalls(int type,int value);
int fwdWildCardCall(int type,int fmt);

typedef struct EQPTask
{
  void (*tskfcn)(void); /* background task function */
  int weight;           /* background task weight factor */
  struct EQPTask *next;
} EQPTASK;
extern EQPTASK *EQPTaskList;

#define EQPBKG void (*)(void)
#define EQPINI void (*)(void)
#define EQPEXI void (*)(void)

#define ALLOCSIZE(b) ((b) + (8 - (b) % 8))

typedef void (*SCCBFCNP)(const char *,const char *,void *);
typedef void (*CYCBFCNP)(int,int,void *);

extern DIEFCNP dieFunction;

#define GLOBALS_KEY_SIZE 132

/**
 * \internal
 * \brief The top level header which defines the incoming legacy data packet.
 *
 */
typedef struct /* struct is passed over the net (legacy) */
{
  UINT16 totalSize; /**< total packet size in bytes (must match the bytes read) */
  char userName[USERNAME_SIZE]; /**< caller name */
  short tineProtocol; /**< tine protocol level (modern: 6 for contracts, 4 for globals) */
  UINT16 number; /**< tine version number (contracts) or number of incoming keywords (globals) */
} PktHdr;
#define PKTHDR_SIZE  22

/**
* \internal
* \brief The top level header which defines the incoming request (or globals) data packet.
*
*/
typedef struct PayloadHeader/* struct is passed over the net (modern, little endian) */
{
  UINT32 totalSizeInBytes;  /**< total packet size in bytes (must match the bytes read) */
  UINT32 headerSizeInBytes; /**< size of this header in bytes (allows for dynamic extension ?) */
  UINT32 pid;               /**< process ID of caller */
  SINT32 tineVersion;       /**< encoded specific version information */
  SINT16 numberSegments;    /**< number of following segments (contract requests, globals keywords) */
  SINT16 tineProtocol;      /**< tine protocol level (modern: 7) :: still at position 18 */
  UINT16 endianness;        /**< endianness of caller : 0 or 1 (little or big) -> all subsequent headers and data */
  UINT16 encoding;          /**< character encoding: ascii, utf-8, uni-code ? -> requires byte swapping */
  char   userName[USERNAME_SIZE]; /**< caller name */
  char   userType[USERTYPE_SIZE]; /**< application type tag : 8 characters */
} ReqHdr, GlbHdr;
#define REQHDR_SIZE  48
#define MINIMUM_REQHDR_PROTOCOL 7

#define SUB_PROT_LOCATION 18

#define PAYLOADHDR_SIZE(p) (p < MINIMUM_REQHDR_PROTOCOL ? PKTHDR_SIZE : REQHDR_SIZE)

/**
* \internal
* \brief The top level header which defines the incoming response data packet.
*
*/
typedef struct
{
  SINT32 tineProtocol; /**< incoming tine protocol as integer -> distinct from legecy response of total bytes */
  UINT32 totalSizeInBytes; /**< total packet size in bytes (must match the bytes read) */
  UINT32 headerSizeInBytes; /**< size of this header in bytes (allows for dynamic extension ?) */
  SINT16 numberSegments; /**< number of following segments (contract responses) */
  UINT8  endianness; /**< endianness of FEC : 0 or 1 (little or big) -> all subsequent headers and data */
  UINT8  encoding;   /**< character encoding: ascii, utf-8, uni-code ? -> requires byte swapping */
  char   fecName[USERNAME_SIZE]; /**< FEC name */
} RspHdr;
#define RSPHDR_SIZE 32

/**
* \internal
* \brief additional information contained in the response packet
*
*/
typedef struct
{
  UINT16 prot; /**< protocol level */
  UINT16 id; /**< subscription id */
  UINT16 cc; /**<  completion code */
  UINT16 fmt; /**< returned format */
  UINT16 stssiz; /**< returned status size */
  UINT32 starttime; /**< subscripton starttime */
  UINT32 sysstamp; /**< system stamp */
  UINT32 userstamp; /**< user stamp */
} RspSpecs;

#define RETURNHDR_SIZE(p) (p < MINIMUM_REQHDR_PROTOCOL ? sizeof(UINT16) : RSPHDR_SIZE)

#define LNKTBL_HASH_SIZE 211

#define REVISIONID(m,r) ((m)*256 + (r))
#define VERSIONID(mj,mn,r) (((mj)<<16)|((mn)<<8)|(r))

#define BO_LITTLE_ENDIAN 0
#define BO_BIG_ENDIAN 1
#define CE_ASCII 0
#define CE_UTF8 1

#define TCBCALL(ce,cc) {\
  ce->isInsideCallback = TRUE;\
  ce->hasReference ?\
  ((XTCBFCNP)ce->cbfcn)((int)ce->cbId,cc,ce->cbRef) :\
  ((TCBFCNP)ce->cbfcn)((int)ce->cbId,cc);\
  ce->isInsideCallback = FALSE;\
}

typedef struct BlackLnkLstStruct
{
  char key[FULL_NAME_MAXSIZE+8];
  int  status;
  struct BlackLnkLstStruct *nxt;
} BLACKLISTED_LINK;
extern BLACKLISTED_LINK *gBlackLnkTbl[LNKTBL_HASH_SIZE];
extern int gBlackLnkTblSize;
BLACKLISTED_LINK *getBlackLnkLstItem(char *ctx, char *srv, char *dev, char *prp);
void appendBlackLnkLst(char *ctx, char *srv, char *dev, char *prp,int status);
void flushBlackLnkLst(void);
void freeConTblEntryMemory(int idx);

#define RELINKED_FALSE  0x00
#define RELINKED_TRUE   0x01
#define RELINKED_NEEDED 0x02
#define RELINKED_STRUCT 0x04
#define RELINKED_BITFIELD 0x08
#define RELINKED_INVALIDDATA 0x10

typedef struct ReLnkCbLstStruct
{
  TCBFCNP cbfcn;
  int lnkId;
  UINT32 cbId;
  void *cbRef;
  short hasReference;
  short isInsideCallback;
  struct ReLnkCbLstStruct *nxt;
} ReLnkCbEntry;

typedef struct ReLnkLstStruct
{
  char key[FULL_NAME_MAXSIZE+8];
  char tag[TAG_NAME_SIZE];
  char field[FIELD_NAME_SIZE];
  int  sts;     /* status code which led to re-linking */
  int  fmt;
  int  siz;
  int  lnkSiz;  /* original link data size */
  int  lnkFmt;  /* original link format */
  BYTE *dat;    /* pointer to original bound data */
  ReLnkCbEntry *cbLst;
  struct ReLnkLstStruct *nxt;
} ReLnkTblEntry;

extern ReLnkTblEntry *gReLnkTbl[LNKTBL_HASH_SIZE];

/**
 * \internal
 * \brief Linked List structure used by clients to manage redirected calls.
 */
typedef struct rdrLnkStruct
{
  char srcContext[CONTEXT_NAME_SIZE];
  char srcServerName[EXPORT_NAME_SIZE];
  char srcProperty[PROPERTY_NAME_SIZE];
  char srcDeviceName[DEVICE_NAME_SIZE];
  char dstContext[CONTEXT_NAME_SIZE];
  char dstServerName[EXPORT_NAME_SIZE];
  char dstDeviceName[DEVICE_NAME_SIZE];
  char dstProperty[PROPERTY_NAME_SIZE];
  int isGroupMember;
  struct rdrLnkStruct *prv;
  struct rdrLnkStruct *nxt;
} RdrTblEntry;

#ifndef REDIR_HASH_SIZE
# define REDIR_HASH_SIZE 509
#endif

extern RdrTblEntry *rdrLnkTbl[REDIR_HASH_SIZE];

RdrTblEntry *appendRedirectedLinkList(char *con,char *srv,char *prp,char *dev,char *dstctx,char *dstsrv,char *dstdev,char *dstprp,int isGrpMem);

/**
 * \internal
 * \brief Linked List structure used by connection table if multiple links to same contract
 */
typedef struct DependentLinkStruct
{
  int linkId;
  int isInactive;
  struct DependentLinkStruct *prv;
  struct DependentLinkStruct *nxt;
} DpdTblEntry;

typedef struct mcaLnkItemStruct
{
  int linkId;
  int boundToId; /* in case of dependency prior to MCA assignment */
  int isCanceled;
  struct mcaLnkItemStruct *nxt;
} McaLnkItem;
/**
 * \internal
 * \brief Linked List structure used by connection table if link is element of multi-channel link
 */
typedef struct mcaLinkStruct
{
  int linkId;
  int isElevatedFormat;
  int ndevices;
  int nlinks;
  McaLnkItem *links;
  char mcaContext[CONTEXT_NAME_SIZE];
  char mcaServerName[EXPORT_NAME_SIZE];
  char mcaProperty[PROPERTY_NAME_SIZE];
  char mcaDevice[DEVICE_NAME_SIZE];
  struct mcaLinkStruct *prv;
  struct mcaLinkStruct *nxt;
} McaTblEntry;
extern McaTblEntry *mcaLnkLst;
#define MCADEV_TAG "mcaDev=>"

typedef struct grpCacheItemStruct
{
  char ctx[CONTEXT_NAME_SIZE];
  char grp[EXPORT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  struct grpCacheItemStruct *nxt;
} GrpCacheItem;
extern GrpCacheItem *gGrpCacheList;
extern time_t gLastGroupCacheAddTime;

void addToGroupCacheList(const char *context,const char *group,const char *server,const char *device);
void flushGroupCacheList(int nitems);

typedef struct
{
  float absolute;
  float percent;
  int offset;
  int length;
  int flags;
} DataToleranceFactor;

#define DTF_NONE     0x00
#define DTF_SUPPRESS 0x01
#define DTF_NOTIFY   0x02

#define GRP_SYNC_NOSYNC 0
#define GRP_SYNC_STARTSYNC 1
#define GRP_SYNC_NOTIFICATION 2
#define GRP_SYNC_CYCLEOFFSET 3
#define GRP_SYNC_INTERVAL 4
#define GRP_SYNC_INSYNC 5

typedef struct grpMemberStruct
{
  int conTblId;
  int glbTblId;
  DTYPE dout;
  char key[STATUS_SIZE+8];
  struct grpMemberStruct *nxt;
} GrpMember;

/**
 * \internal
 * \brief Linked List structure used by clients to manage grouped calls.
 */
typedef struct grpStruct
{
  short numPending; /**< number of group events pending */
  short numInGrp; /**< number of elements in group */
  short grpCc; /**< group completion code */
  short grpSyncLevel; /**< group synchronization level */
  short grpIncMono; /**< group count increments monotonically */
  short grpHeadInc; /**< group head count has incremented */
  int grpUpdateInterval; /**< group update interval */
  int grpBndWdthT; /**< group time dispersion */
  int grpBndWdthC; /**< group cycle dispersion */
  int grpSysStamp; /**< assigned group system stamp (cycle number) */
  int grpLastStamp; /**< last group system stamp (cycle number) */
  int grpHeadLnkId; /**< designated group master's connection table id */
  struct timeval grpTimeUpdate; /**< time of this update */
  struct timeval grpLastUpdate; /**< time of last update */
  size_t grpId; /**< group identification (cbFcn address) */
  GrpMember *members; /**< group member list */
  struct grpStruct *next; /**< next entry in list */
} GrpTblEntry;

#define WILDCARD_LINKID(i) ((i) & 0xffff)
#define WILDCARD_LISTID(i) ((i) >> 16)
#define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<16))

/**
 * \internal
 * \brief Table entry structure used by clients to manage wildcard calls
 */
typedef struct wildCardTableStruct
{
  char key[FULL_NAME_SIZE];
  int linkId;  /* xref into connTableEntry (parent link) */
  int cbId;    /* user's callback id */
  int length;  /* length of device list OR GENS member list */
  int mode;    /* link mode */
  int numPending; /* number waiting to notify */
  int numActive;  /* number of active links in this set */
  BYTE hasReference; /* user's callback prototype flag */
  BYTE hasGensParent; /* if the parent link is to the GENS */
  BYTE isInsideCallback; /* flag */
  BYTE active; /* this entry is currently active */
  NAME64 *list;/* list of device names OR GENS member servers */
  BYTE *data;  /* either floats or longs */
  int *status; /* list of individual comp codes */
  int *listId; /* list of individual link ids */
  DTYPE dout;  /* caller's output */
  DTYPE din;   /* caller's input */
  TCBFCNP cbfcn; /* user's callback */
  void *cbRef; /* user's callback reference parameter */
  struct wildCardTableStruct *nxt; /* next pointer */
} WcTblEntry;

extern WcTblEntry *wcTable[LNKTBL_HASH_SIZE];
extern int nWildCardTableEntries;

typedef struct cbXRefTblStruct
{
  int linkId;  /* xref into connTableEntry */
  int cbId; /* user's callback id */
  TCBFCNP cbfcn; /* user's callback */
  short hasReference; /* user's callback prototype flag */
  short isInsideCallback;
  void *cbRef; /* user's callback reference parameter */
} cbXRefTblEntry;

extern cbXRefTblEntry **cbXRefTbl;
extern int nCbXRefTblEntries;

#define UNASSIGNED_INDEX -1
#define UNASSIGNED_CALLBACKID 0xeeeeeeee
#ifndef SMALL_TINE_LIB
# define DEFAULT_RECV_QUEUE_DEPTH 10
#else
# define DEFAULT_RECV_QUEUE_DEPTH 1
#endif
typedef struct ConnTableQueueStruct
{
  BYTE *dataOut;
  int dataSize;
  int dataStatus;
  int statusCode;
  UINT32 timeStamp;
  int timeStampUSEC;
  int usrStamp;
  int sysStamp;
  int xfrReason;
  struct ConnTableQueueStruct *next;
} ConTblQueue;

#define LWD_POLLING_RATE 1000
#define LWD_PROPERTY "SRVSTARTTIME"
#define LWD_DTAG "WDOG"
#define LWD_TIMER_DEADBAND 5
typedef struct LnkWatchDogStruct
{
  int lnkId;
  int lnkLastStatus;
  int nlinks;
  UINT32 srvStartTime;
  struct LnkWatchDogStruct *nxt;
} LnkWatchDog;
extern LnkWatchDog *wdLinks;
extern int lwdPollingInterval;
/**
 * \internal
 * \brief Connection table structure used by clients.
 *  structure is local (not transported over the network)
 *  => can change as needed
 *  No exported routines use ConTblEntry !
 */
typedef struct
{
  char expName[EXPORT_NAME_SIZE];/**< eqm server export name */
  char prpName[PROPERTY_NAME_SIZE];/**< eqm property */
  char devName[DEVICE_NAME_SIZE];/**< eqm device name */
  char ctxName[CONTEXT_NAME_SIZE];/**< eqm context */
  char *extStringspace; /**< eqm extended string space (long device names) */
  short extStringsize;  /**< eqm extended string space size */
  short access;         /**< data access */
  short formatIn;       /**< eqm data format down */
  short formatOut;      /**< eqm data format up */
  UINT32 sizeIn;        /**< data array size down */
  UINT32 sizeOut;       /**< data array size up */
  UINT32 sizeBytesOut;  /**< size bytes of output request */
  char strTagIn[TAG_NAME_SIZE];/**< eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE];/**< eqm structure tag up */
  char key[FULL_NAME_SIZE]; /**< link table key for this entry */
  short dataStatus;     /**< signal new arrival (negative completion code) */
  short tocounter;      /**< link timeout counter */
  int  srvIdx;          /**< index into Srv Table */
  int  fecIdx;          /**< index into FEC database */
  UINT32 pollingInterval;/**< polling interval for connectioned clients */
  UINT32 heartbeat;     /**< used by subscriptions */
  UINT16 mode;          /**< used by subscriptions */
  UINT16 xferReason;    /**< transfer flag applicable to last data transfer */
  int tineProtocol;     /**< tine protocol level */
  int inetProtocol;     /**< transport protocol (IPX,SPX,UDP,TCP) */
  UINT32 timeLastSync;  /**< last time sync - request or response or connection check */
  UINT32 timeMsecLastSync;/**< last time sync - request or response or connection check (msec fraction) */
  UINT32 starttime;     /**< entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /**< arriving TINE timestamp */
  int dtimestampUSEC;   /**< arriving TINE timestamp (usec part) */
  int datastamp;        /**< call applied data stamp */
  int systemstamp;      /**< system applied data stamp */
  UINT32 lastDataSize;  /**< arriving data size */
  UINT32 timemark;      /**< a UTC time mark for internals */
  SOCKET tcpSck;        /**< tcp/ip socket if connected */
  INADR mcastGrp;       /**< the multicast group ip addr (if attached to a group) */
  int  counter;         /**< current subscription counter */
  int  renewalMultiplier;/**< 'learned' renewal multiplier to establish maximum counter */
  char *dataIn;         /**< pointer to associated (serialized) data down */
  char *dataPtr;        /**< pointer (reference) to associated data up */
  char *dataMarshalled; /**< pointer to marshalled data up */
  char *bufOut;         /**< pointer to raw (serialized) data up buffer if necessary (payload > 1 MTU) */
  BYTE *bufOutBlks;     /**< pointer to data-up buffer block progress if necessary */
  char *bufCmp;         /**< pointer to comparison buffer (serialized) if suppressHearbeatNotification == TRUE */
  char *bufVtl;         /**< pointer to volatile buffer if volatile format (e.g. CF_STRING) */
  DataToleranceFactor *dtf;/**< pointer to comparison tolerance if applied */
  UINT32 allocatedBytes;/**< if dataOut has allocated memory (instead of bound data) */
  UINT32 numblks;       /**< progress counters for incoming data */
  UINT32 blksin;        /**< progress counters for incoming data */
  UINT32 bytesin;       /**< progress counters for incoming data  */
  UINT32 nextDataSize;  /**< next outgoing data size in bytes */
  UINT32 nextDataPtr;   /**< next outgoing data array pointer */
  UINT32 bytesRemaining;/**< number of bytes data remaining */
  UINT32 bytesTotal;    /**< total number of bytes data */
  UINT32 datamtu;       /**< outgoing data minimum transport unit */
  UINT32 blkidIn;       /**< block Id of data to send */
  UINT32 numblksIn;     /**< num blks input data to send */
  UINT32 blknumIn;      /**< this blk num of input data */
  UINT32 cbId;          /**< pivate callback ID */
  TCBFCNP cbfcn;        /**< callback function */
  UINT8 isGrouped;      /**< flag to indicate if element is grouped */
  UINT8 notifyPending;  /**< flag to indicate if notification is pending */
  UINT8 suppressCallback;/**< flag to withhold callback notification */
  UINT8 relinked;       /**< flag to indicate if entry is relinked (or action needed) */
  UINT16 subId;         /**< server subscription ID (applicable in MCAST) */
  UINT32 rdrId;         /**< if non-null gives child connection ID for redirected calls */
  WcTblEntry *wcLink;   /**< wildcard entry if relevant */
  short linkStatus;     /**< current link status */
  short lastStatusReported; /**< status in last callback */
  short rbModFlags;     /**< readback modifcation flags (e.g. RB_USE_ERRORVALUE, RB_USE_TIMESTAMP) */
  short needsNotification;/**< flag to call NotifyClient() or not */
  short hasNotifiedOnce; /**< necessary for addressing nofication suppression issues */
  short hasUserCallbackId;/**< flag : TRUE if set by caller, otherwise false */
  short lastLnkErrSrc;  /**< LCL or RMT */
  short isWildcardLink; /**< parent link is a wildcard call */
  short isWaitingSyncNotification; /**< for synchronous calls */
  short blcounter;      /**< blacklist counter */
  short formatReq;      /**< original requested format (prior to any re-linking) */
  short reservedWord;   /**< preserve an even number of shorts */
  UINT32 sizeReq;       /**< original requested data size (prior to any re-linking) */
  char lastStatusString[STATUS_SIZE]; /**< temporary storage for return status */
  BYTE  errValue[64];   /**< buffer for assigned error value */
  int cancelPending;    /**< delay CANCEL because of dependencies */
  int hasAcknowledged;  /**< flag to smooth subscription renewal transitions */
  int linkId;           /**< this table entry index */
  int boundToId;        /**< id of parent link if this one is bound (>0) */
  int glbTblId;         /**< globals table id if this refers to a global */
  int mcaIndex;         /**< multi-channel array index (if any) */
  McaTblEntry *mcaEntry;/**< pointer to multi-channel entry (if this link is an Mca parent) */
  McaTblEntry *mcaLink; /**< pointer to multi-channel link (if any) */
  DpdTblEntry *dpdLink; /**< pointer to dependency list (if this link is a parent) */
  ReLnkTblEntry *rlkEntry;/**< pointer to relink list (if any) */
  RdrTblEntry *rdrLink; /**< pointer into the redirection table */
  int grpLatency;       /**< notification lag of this group element (if grouped) */
  int syncOffset;       /**< learned offset in sysStamp of this group element (if grouped) */
  GrpTblEntry *grp;     /**< pointer to the group table entry (if relevant) */
  int queueLen;         /**< current link queue length */
  int queueDepth;       /**< current link queue depth */
  ConTblQueue *queue;   /**< current link queue */
  int idleCount;        /**< idle count of a listening link */
  int renewalThreshold; /**< if > 0 gives the subscription counter at which a renewal request is issued */
  short hasReference;   /**< TRUE => use protocol with cteRef ! */
  short isInsideCallback;/**< flag indicating link entry is inside callback routine */
  short cbDelay;        /**< registered number of cb cycles to delay callback notification */
  short cbDelayCounter; /**< counter for delayed callback notification */
  void *cbRef;          /**< reference pointer passed in AttachLinkEx2() */
  LnkWatchDog *wdLink;  /**< link watchdog reference if relevant */
} ConTblEntry;

#define ALLOWED_IDLE_LISTENING_COUNT 300

/**
 * \internal
 * \brief Connection table parameters
 *  structure is query-able (available over the network)
 *  => changing this can break things !
 */
typedef struct /* net */
{
  char expName[EXPORT_NAME_SIZE]; /* eqm function name */
  char prpName[PROPERTY_NAME_SIZE]; /* eqm property */
  char devName[DEVICE_NAME_SIZE]; /* eqm device name */
  char context[CONTEXT_NAME_SIZE]; /* eqm context */
  char eqm[EQM_NAME_SIZE];
  UINT16 mode;          /* used by subscriptions */
  short access;         /* data access */
  short formatIn;       /* eqp data format down */
  short formatOut;      /* eqp data format up */
  UINT32 sizeIn;        /* data array size down */
  UINT32 sizeOut;       /* data array size up */
  char strTagIn[TAG_NAME_SIZE]; /* eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE]; /* eqm structure tag up */
  UINT32 pollingInterval; /* polling rate for connectioned clients */
  UINT32 heartbeat;     /* used by subscriptions */
  UINT32 starttime;     /* entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /* arriving TINE timestamp */
  int dtimestampMSEC;   /* arriving TINE timestamp (msec fraction) */
  int  counter;         /* current subscription counter */
  short linkStatus;     /* current link status */
  short reserved;
} ConTblInfo;
#define CONTBL_QUERY_SIZE sizeof(ConTblInfo)

/* GetConnectionTable */
/**
 * \brief Gets the current connection table.
 *
 * A client's connections are managed and maintained in a connection table.  You can retrieve
 * the current connection table by using this call
 *
 * \param tbl is a reference to block of memory which will hold the ConTblInfo array
 * giving the current client connection table.
 * \param tblSize (input) is a pointer to an integer value containing the maximum size (number of entries)
 * the the tbl reference can hold. (output) will contain the size of the current connection table.
 *
 * \return 0 upon success or a TINE return code.
 *
 * \see GetConnectionTableCapacity(), SetConnectionTableCapacity()
 */
TINE_EXPORT int GetConnectionTable(ConTblInfo *tbl,int *tblSize);

/* GetConnectionList */
/**
 * \brief Gets the current connection table.
 *
 * A client's connections are managed and maintained in a connection table.  You can retrieve
 * the current connection table by using this call
 *
 * \param lstbuf is a reference to a string buffer which will hold the connection table list
 * \param bufsize gives the size in bytes of the string buffer passed in the first argument.
 *
 * \return 0 upon success or a TINE return code.
 *
 */
TINE_EXPORT int GetConnectionList(char *lstbuf,int bufsize);

#define EMPTY_PROP    0
#define TARGET_PROP   1
#define WILDCARD_PROP 2
#define STOCK_PROP    3

#ifndef PROPERTY_HASH_SIZE
# define PROPERTY_HASH_SIZE 211
#endif
#define PROPERTY_HASH_INDEX(i) ((i) & 0xff)
#define PROPERTY_NUM_INDEX(i) (((i) >> 8) & 0xffff)
#define MAKE_PROPERTY_ID(i,indx) (((i) << 8) | (indx))

#define PRP_REDIR_CTX_OFFSET 0
#define PRP_REDIR_SRV_OFFSET 32
#define PRP_REDIR_DEV_OFFSET 64
#define PRP_REDIR_PRP_OFFSET 128

#define PRP_REDIR_SRV_OFFSET_P5 0
#define PRP_REDIR_PRP_OFFSET_P5 16

int assertRedirectionValid(char *tag,char *eqm,char *rdr);

#define STRICTLY_ILLEGALCHARS "[^\"*&]"
#define RUDECHARS "[^\\\\/\n\t ]+"
int assertNameIsValid(const char *name,const char *badchars,int length);

#ifndef DEVICES_HASH_SIZE
# define DEVICES_HASH_SIZE 211
#endif
#define DEVICES_HASH_PAYOFF (DEVICES_HASH_SIZE/10) /* don't hash if fewer modules than this */

typedef struct HashXRef
{
  int index;
  struct HashXRef *next;
} HashXRefStruct;

#ifndef STOCKPROP_HASH_SIZE
# define STOCKPROP_HASH_SIZE 103
#endif

extern HashXRefStruct **StockPropXRefTable;

typedef struct
{
  NAME64 *prpList;
  int prpListSize;
  char filename[64]; /* used as a 'tag' */
} EqmDevicePropertyList;

/**
 * \internal
 * \brief export info struct (to handle file input)
 *
 */
typedef struct ExpInfoStruct
{
  char EqmContext[CONTEXT_NAME_SIZE];    /**< exported equipment name context */
  char ExpName[EXPORT_NAME_SIZE];        /**< exported equipment module name */
  char SubSystem[SUBSYSTEM_NAME_SIZE];   /**< equipment module sub system */
  char EqmName[EQM_NAME_SIZE];           /**< local equipment module name */
  char EqmGroup[EXPORT_NAME_SIZE];       /**< group name, if applicable */
  char EqmGroupDevPrefix[XFIX_SIZE];     /**< group device prefix, if applicable */
  char EqmGroupDevPostfix[XFIX_SIZE];    /**< group device postfix, if applicable */
  char EqmMaster[EXPORT_NAME_SIZE];      /**< eqm master name, if applicable */
  char EqmSlaveMaster[EXPORT_NAME_SIZE]; /**< eqm master name for slave operation, if applicable */
  char PrpName[PROPERTY_NAME_SIZE];      /**< registered property name */
  char fmtOut[64];                       /**< output format string (possibly decorated) */
  char fmtIn[64];                        /**< input format string (possibly decorated) */
  char description[128];                 /**< description (possibly parseble equ info) */
  char redirection[DEVICE_REDIR_SIZE];   /**< redirection information */
  char access[ACCESS_STR_SIZE];          /**< access string */
  char units[UNITS_SIZE];                /**< nominal engineering units */
  char xunits[UNITS_SIZE];               /**< x-axis units (for double arrays) */
  char region[TAG_NAME_SIZE];            /**< default alarrm region tag for device instances */
  int  PrpSizeOut;                       /**< output data size */
  int  PrpSizeIn;                        /**< input data size */
  int  PrpId;                            /**< associated property id */
  int  EqmGroupIdx;                      /**< group index if applicable */
  int  DevSpace;                         /**< total amount of space initially allocated */
  int  altDevSpace;                      /**< total amount of space initially allocated (using NUM_MODULES) */
  int  deadband;                         /**< access deadband in milliseconds */
  float max;                             /**< nominal maximum value */
  float min;                             /**< nominal minimim value */
  float xmax;                            /**< nominal x-axis maximum value (double arrays) */
  float xmin;                            /**< nominal x-axis minimum value (double arrays) */
} XIS;

typedef struct ContractResourceStruct
{
  DTYPE din;
  DTYPE dout;
  char devName[DEVICE_NAME_MAXSIZE];
  char devProp[PROPERTY_NAME_SIZE+4];
  short access;
} CRS;

typedef struct
{
  char *rdrName;
} EqmDeviceRdrListType;

#define EQMCALL(ph,dn,pn,od,id,ax,rf) (\
  ph == NULL ?\
    non_existent_function :\
    rf != NULL ?\
      ((XEQMFCNP)ph)(dn,pn,od,id,ax,rf) :\
      ((EQMFCNP)ph)(dn,pn,od,id,ax)\
)


#define TSKCALL(el,tsk) (\
  el->hasReference ?\
  ((XEQMTSKP)el->tsk)(el->EqmRef) :\
  ((EQMTSKP)el->tsk)()\
)

typedef struct
{
  BYTE IpxNet[4];
  BYTE IpxNode[6];
  BYTE IpxImmAdr[6];
} IpxAdr;
typedef struct
{
  int sckRmt;
  int sckLcl;
} IpxSck;
typedef union
{ /* 16-bytes */
  BYTE ipv6[16];
  BYTE ipv4[4];
  IpxAdr ipx;
} FecNetAdr;
typedef struct
{
  int sck;
  int err;
} PipeAdr;
typedef union
{ /* force 8-byte size */
  HWND wndMmf;
  PipeAdr pipe;
  BYTE mac[6];
  BYTE reserved[8];
} FecLclAdr;
typedef union
{ /* force 8-byte size */
  HWND wndMmf;
  PipeAdr pipe;
  SOCKET sck;
  IpxSck ipx;
  BYTE reserved[8];
} EpSck; /* Endpoint socket */
typedef struct
{ /* force 16-byte size */
  SCKADR ip;
  IpxAdr ipx;
  HWND wndMmf;
  UINT32 prot;
  BYTE reserved[16];
} EpAdr; /* Endpoint address */

/**
 * \internal
 * \brief Client information struct
 */
typedef struct
{
  char    usr[USERNAME_SIZE]; /**< user name of client */
  EpAdr   adr;         /**< end point address of client */
  EpSck   ep;          /**< communication end point socket */
  char    typ[USERTYPE_SIZE]; /**< application type string of client */
  int     pid;         /**< process id of client application */
  short   ncontracts;  /**< number of active contract subscriptions */
  short   tineProtocol;/**< target client protocol level */
  short   inetProtocol;/**< transport protocol */
  short   reserved;    /**< round out the struct definition */
} ClnHdr;
#define CLIENT_SIZE sizeof(ClnHdr)
#define CLNHDR_CMP_SIZE (USERNAME_SIZE + sizeof(EpAdr))
extern ClnHdr gLastAllowedCln;
void getPeerAddrStr(char *addr,ClnHdr *cln);

/**
 * \internal
 * \brief Definitions relevant to an Access Lock
 *
 */
#define LOCK_UNLOCKED 0
#define LOCK_PREEMPTIVE 1
#define LOCK_PERSISTENT 2
#define LOCK_ABORT 3
#define LOCK_XREAD 0x100
#define LOCK_NOCALLBACKS 0x200
#define MAX_LOCK_DURATION 60

/**
 * \internal
 * \brief AccessLock used on the server side
 *
 */
typedef struct
{
  ClnHdr cln;
  int lockType;
  int lockDuration;
  UINT32 lockToken;
  UINT32 lockStart;
} AccessLock;

/**
 * \internal
 * \brief AccessLock used on the client side (link-listed)
 *
 */
typedef struct AccessLockListItemStruct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  int lockType;
  int lockDuration;
  int lockLinkId;
  int lockLinkStatus;
  int lastSent;
  struct AccessLockListItemStruct *nxt;
} AccessLockListItem;

extern AccessLockListItem *accessLockLst;
int rmvAccessLock(char *context, char *server,int wait);
int _setAccessLock(char *context, char *server, int lockType,int lockDuration);
AccessLockListItem *getAccessLockItem(char *context, char *server);
void checkAccessLockItems(void);
int GetDeviceAclInfo(char *eqm,int devnr);

/**
 * \internal
 * \brief property deadband access item used on server side
 *
 */
typedef struct PrpDbaItem
{
  char eqm[EQM_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char usr[USERNAME_SIZE];
  struct timeval lastAccess;
  SCKADR addr;
  int deadband;
  int access;
  struct PrpDbaItem *nxt;
} PrpDbaItem;

extern PrpDbaItem *prpDbaLst;
PrpDbaItem *getPrpDbaItem(char *eqm,char *prp);

/**
 * \internal
 * \brief user groups registry (determined from net or local groups) (link-listed)
 *
 * used for security access at the server side
 */
typedef struct UsrGrpsLstStruct
{
  char name[GROUPNAME_SIZE];
  char type[GROUPSOURCE_SIZE];
  NAME16 *members;
  int nmembers;
  struct UsrGrpsLstStruct *nxt;
} UsrGrpsLst;
extern UsrGrpsLst *gUsrGrps;
extern int nUsrGrps;

TINE_EXPORT int GetUserGroupMembers(char *filename,char *grp,char *src,NAME16 **members,int *nmembers);

typedef struct
{
  char dev[DEVICE_NAME_SIZE];
  char reg[TAG_NAME_SIZE];
  int code;
} DeviceRegionItem;
int GetValidDeviceRegion(char *eqm,char *regionTag);
int GetRegionForDevice(char *eqm,char *regionTag,char *devName);
int getRegionRules(char *eqm);
char *GetRegionTag(char *eqm,int code);

#define DEV_REGION_CENTRAL 0
#define DEV_REGION_UNKNOWN 255

typedef struct
{
  char grpName[DEVICE_NAME_SIZE];
  int grpSize;
  int grpIndex;
} DeviceMcaGroup;

#define DEVICE_OFFLINE 0x80000000
typedef struct ExportDeviceStruct
{
  char *name;                 /**< pointer to device name */
  char *desc;                 /**< pointer to device description (64-char) */
  char *rdr;                  /**< pointer to device redirection string (132-char) */
  char *loc;                  /**< pointer to device location (64-char) */
  EqmDevicePropertyList *prps;/**< pointer to properties supported by this device */
  AclInfoType *aclLst;        /**< pointer to device specific acls */
  DeviceMcaGroup *mcaGrp;     /**< pointer to device MCA group element */
  ALARM *almLst;              /**< associated alarm list */
  float zpos;                 /**< device z position */
  UINT32 mask;                /**< device mask (doocs sys_mask) & offline bit */
  int region;                 /**< specific device region setting */
  int number;                 /**< assigned device number (slot in device table) */
  int isReserved;             /**< if TRUE then entry is just a place holder */
} ExportDeviceType;

int makeEqmDevLst(ExportDeviceType **devlst,int len);
void freeEqmDevLst(ExportDeviceType *devlst,int len);
void freeEqmDevType(ExportDeviceType *devlst, int idx, int len);

/** Device info structure used in queries */
typedef struct DeviceInfoStruct
{
  char ctx[CONTEXT_NAME_SIZE]; /**< context of device name */
  char srv[EXPORT_NAME_SIZE];  /**< server of device name */
  char name[DEVICE_NAME_SIZE]; /**< device name */
  char desc[DEVICE_DESC_SIZE]; /**< device description (64-char) */
  char loc[DEVICE_LOCATION_SIZE];/**< device location (64-char) */
  float zpos;                  /**< device z position */
  UINT32 mask;                 /**< device mask (doocs sys_mask) & offline bit */
  int region;                  /**< specific device region setting */
  int number;                  /**< assigned device number (slot in device table) */
} DeviceInfoStruct;

#define DEVQSR5_TAG "DEVQSr5"

typedef struct
{
  char name[64];
  int num;
  int msk;
  int rgn;
  float zpos;
  char dsc[64];
  char loc[64];
} DEVMF;
int GetDevicesFromManifest(char *eqm, char *prp, DEVMF **devs, int *listSize);
typedef struct
{
  char name[64];
  char dsc[64];
  char alias[64];
  char memfn[128];
  int id;
} PRPMF;
int GetPropertiesFromManifest(char *eqm, PRPMF **prps, int *listSize);
/**
 * \brief Linked list structure used to hold equipment module information.
 *
 */
typedef struct ExportListTag /* local */
{
  char   EqmContext[CONTEXT_NAME_SIZE]; /**< exported equipment name context */
  char   EqmExportName[EXPORT_NAME_SIZE]; /**< exported equipment module name */
  char   EqmSubSystem[SUBSYSTEM_NAME_SIZE]; /**< equipment module sub system */
  char   EqmName[EQM_NAME_SIZE]; /**< local equipment module name */
  char   EqmAlias[EQM_NAME_SIZE]; /**< local equipment alias name */
  char   EqmGroup[EXPORT_NAME_SIZE]; /**< group name, if applicable */
  char   EqmGroupDevPrefix[XFIX_SIZE]; /**< group device prefix, if applicable */
  char   EqmGroupDevPostfix[XFIX_SIZE]; /**< group device postfix, if applicable */
  char   EqmRegion[TAG_NAME_SIZE]; /**< default region tag */
  int    groupIndex;        /**< group index if applicable */
  int    exportId;          /**< local ID */
  ExportPropertyListStruct  /**< pointer to chained equipment property hash list */
         *EqmPrpLst[PROPERTY_HASH_SIZE+1]; /**< Add room for Wild-card list */
  int    (*EqmPropertyQueryFcn)(char *devName,PrpQueryStruct **prpqs);
  int    (*XEqmPropertyQueryFcn)(char *devName,XPropertyQueryStruct **xpqs);
  short  hasPropertyDeviceNames; /**< TRUE if a property sepcific device list is known */
  short  hasRegisteredDevicesNames; /**< TRUE if a property sepcific device list is registered via API */
  int    EqmWcPrpLstCapacity;/**< size of list buffer space used in eqm wildcard queries */
  NAME64 *EqmWcPrpLst;      /**< pointer to property list buffer space used in eqm wildcard queries */
  ExportDeviceType *EqmDevLst;/**< pointer to device name list */
  HashXRefStruct **EqmDevXRef; /**< ModuleName cross-reference (DEVICES_HASH_SIZE in length) */
  short  EqmNumProperties;  /**< total number of properties supported */
  short  EqmNumDevices;     /**< total number of devices supported */
  short  EqmDevLstCapacity; /**< total amount of space initially allocated */
  short  hasDevQry;         /**< boolean flag for device-query precedence */
  short  registered;        /**< registered with name server */
  short  region;            /**< default region setting for device instances */
  int nRegionItems;         /**< size of acquired region patterns list */
  DeviceRegionItem *regionsLst;/**< acquired region patterns for device instances */
  ADSList *adsTable;        /**< alarm definition structure table (linked List) */
  short  adsTableSize;      /**< number of alarm definitions */
  short  alarmSystem;       /**< alarm system flag */
  AMSDynSet amsDynSet;      /**< brief alarm message containing totals */
  AlmInfo almInfo;          /**< alarm snapshot information */
  short  nclients;          /**< number of active clients */
  short  ncontracts;        /**< number of active contracts */
  short  nfollowers;        /**< number of known possible clients, i.e. 'followers' */
  short  capfollowers;      /**< capacity of followers list */
  NAME64 *followers;        /**< list of known possible clients (followers) */
  int hasPropertyHandler;   /**< true if any property handler or equipment function registered */
  EQMFCNP eqmFcn;           /**< associated equipment function */
  EQMTSKP eqmBkg;           /**< associated background task */
  THRHANDLE hEqmBkg;        /**< handle to background task (in case of threads) */
  SINT32 rateBkg;           /**< background polling rate msec */
  short bkgbusy;            /**< background busy semaphore */
  short inidone;            /**< init task done flag */
  short hstdone;            /**< history init done flag */
  short almdone;            /**< alarm init done flag */
  short isidle;             /**< eqm + tasks are in an idle state */
  short isbusy;             /**< eqm fcn call is busy with WRITE call */
  short busycnt;            /**< busy counter for 'not_ready' */
  BYTE isinside;            /**< eqm fcn is being accessed */
  BYTE isterminating;       /**< termination signal */
  struct timeval lastBkg;   /**< background task last polled */
  EQMTSKP eqmIni;           /**< associated initialization  */
  EQMTSKP eqmExi;           /**< associated cleanup */
  SCCBFCNP staChgCb;        /**< possible state-change callback */
  void *staChgRef;          /**< possible state-change callback reference */
  char EqmLastErrorString[STATUS_SIZE];
  int EqmLastErrorCode;     /**< last EQM access status */
  int EqmNumCalls;          /**< number of times this EQM has been called since startup */
# ifdef MULTITHREADED
  CRS EqmCrs;               /**< ContractResourceStruct (mutex protected) */
# ifdef UNIX
  pthread_mutex_t EqmBusyMutex; /**< EQM primary mutex (posix builds) */
  pthread_mutex_t DataProtectionMutex; /**< EQM mutex for dispatch and background synchronization (posix builds) */
# endif
  MXHANDLE hEqmBusyMutex;    /**< handle to EQM primary mutex */
  MXHANDLE hDataProtectionMutex; /**< handle to EQM mutex for dispatch and background synchronization */
# else
  int hEqmBusyMutex;         /**< handle to EQM primary pseudo-mutex for non-multithreaded builds */
# endif
  int hasReference;         /**< TRUE => use prototypes with EqmRef ! */
  void *EqmRef;             /**< reference pointer to use in extended dispatch functions */
  time_t eqmStartupTime;    /**< eqm specific startup time */
  time_t eqmBaseline;       /**< eqm baseline */
  char usersFileName[32];   /**< either users.csv or <eqm>/users.csv */
  char netsFileName[32];    /**< either ipnets.csv or <eqm>/ipnets.csv */
  AclInfoType aclLst;       /**< eqm specific acls */
  int chkAcl;               /**< check users ACL bits (EQM_CHECK_USR, EQM_CHECK_GRP, EQM_CHECK_NET, etc. */
  int usersScanFlag;        /**< -1 => don't scan, 1 => scan in FEC_HOME, 2 => scan in <eqm> dir */
  int netsScanFlag;         /**< -1 => don't scan, 1 => scan in FEC_HOME, 2 => scan in <eqm> dir */
  int failoverType;         /**< _NONE, _MASTER, or _SLAVE */
  int failoverState;        /**< _NONE, _MASTER, or _SLAVE - this is the current state*/
  char EqmMaster[EXPORT_NAME_SIZE]; /**< applicable if master or slave */
  char EqmSlaveMaster[EXPORT_NAME_SIZE]; /**< applicable if slave */
  AccessLock accessLock;    /**< possible access lock */
  struct ExportListTag *next; /**< Linked list */
} ExportListStruct;

#define HAS_PRPQRYFCN(e) ((e)->EqmPropertyQueryFcn != NULL || (e)->XEqmPropertyQueryFcn != NULL)
extern int gEqmFcnAccessor;
extern ExportPropertyListStruct *ExportPropertyList;
extern ExportListStruct *ExportList;
extern int gEqmListSize;
extern int gSystemUseDataProtection;
extern int gClientIsClosing;
extern int gUseMmfJumbos;
int getNumReservedDevices(ExportListStruct *el);

/**
 * \internal
 * \brief A Structure which is used to hold save-and-restore information at initialization
 *
 * run thru during initial pass of serverCycle()
 */
typedef struct srLstStruct
{
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  int restored;
  int uninitialized;
  int status;
  time_t ts;
  struct srLstStruct *nxt;
} srLst;
extern srLst *gSrLst;
extern time_t lastSaveRestoreFileTime;
void applyStoredPrpValues(ExportListStruct *el);
void writeSrValues(ExportListStruct *el,char *dev,char *prp,DTYPE *din);
void addSrPrp(ExportListStruct *el,ExportPropertyListStruct *pl,int status);
void freeSrPrpList(void);
char *getDevNameAtIndex(ExportListStruct *el,ExportPropertyListStruct *pl,int idx);
srLst *getSaveRestoreItem(ExportListStruct *el,ExportPropertyListStruct *pl);
int acqDefPrpValsFromEqm(char *eqmName,char *prpName,BYTE *data,int num,int devNr);
int applyRegionRules(ExportListStruct *el);

int getFollowersFromFile(char *eqm);
int writeFollowersToFile(char *eqm);
int addFollower(char *eqm,char *follower,int length);
int rmvFollower(char *eqm,char *follower);
int dumpFollowerList(char *eqm);
extern int gIgnoreFollowers;
TINE_EXPORT void SetIgnoreFollowers(int value);
TINE_EXPORT int ApplyStoredPropertyValues(const char *eqm,const char *property);

/**
 * \internal
 * \brief legacy structure which is used to map a EqmContext/ExportName to a FecName/EqmName.
 *
 * The client uses EqmContext and Export name to address the FEC.
 * TINE internally uses the FecName and the Equipment name.
 *
 * The TINE library uses this structure for internal caches.
 *
 * \note struct is passed over the net (legacy: release 3.xx)
 *
 */
typedef struct
{
  char  FecName[FEC_NAME_SIZE];              /**< FEC Name */
  char  SubSystem[SUBSYSTEM_NAME_SHORTSIZE]; /**< short sub-system name */
  char  ExportName[EXPORT_NAME_SHORTSIZE];	 /**< Export Name (legacy size) */
  char  EqmName[EQM_NAME_SHORTSIZE+1];	     /**< Equipment name (legacy size) */
  char  EqmContext[CONTEXT_NAME_SHORTSIZE];	 /**< EqmContext (legacy size) */
  short alarmSystem;                         /**< alarm system flag (never used) */
} RPCDataStruct;
#define RPCDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SHORTSIZE+EXPORT_NAME_SHORTSIZE+EQM_NAME_SHORTSIZE+1+CONTEXT_NAME_SHORTSIZE+2)

/**
 * \internal
 * \brief structure which is used to map a EqmContext/ExportName to a FecName/EqmName.
 *
 * The client uses EqmContext and Export name to address the FEC.
 * TINE internally uses the FecName and the Equipment name.
 *
 * The TINE library uses this structure for internal caches.
 * Sent to and received from the ENS
 *
 * \note struct is passed over the net (104 bytes)
 *
 */
typedef struct
{
  char  FecName[FEC_NAME_SIZE];         /**< FEC Name */
  char  SubSystem[SUBSYSTEM_NAME_SIZE]; /**< sub-system name */
  char  ExportName[EXPORT_NAME_SIZE];   /**< Export Name */
  char  EqmName[EQM_NAME_SIZE];         /**< Equipment name */
  char  EqmContext[CONTEXT_NAME_SIZE];  /**< EqmContext */
} ExpDataStruct;
#define EXPDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SIZE+EXPORT_NAME_SIZE+EQM_NAME_SIZE+CONTEXT_NAME_SIZE)

/**
 * \internal
 * \brief legacy structure which contains the FEC address information
 *
 * The TINE library uses this structure for internal caches.
 *
 * \note struct is passed over the net (legacy: release 3.3xx)
 *
 */
typedef struct
{
  char fecName[FEC_NAME_SIZE];
  BYTE reserved[3];
  BYTE Port[2];
  BYTE Network[4];
  BYTE Node[6];
  BYTE ImmediateAddress[6];
  char IP[16];
  BYTE portOffset;
  char IPh_addr[4];
  short  Prot;
} RPCFecStruct;
#define RPCFECSTRUCT_SIZE (FEC_NAME_SIZE+3+2+16+21+2)

/* these macro re-definitions of the IPX address fields will make sections of the code easier to read */
#define MMFhwnd IPXImmediateAddress
#define PIPEsck IPXNetwork
#define PIPEerr IPXNode[0]
/**
 * \internal
 * \brief structure which contains the FEC address information
 *
 * The TINE library uses this structure for internal caches.
 * Sent to and received from the ENS
 *
 * \note struct is passed over the net (64 bytes)
 *
 */
typedef struct
{
  char fecName[FEC_NAME_SIZE]; /**< FEC name */
  BYTE IPXNetwork[4];          /**< IPX addr info (or internal) */
  BYTE IPXNode[6];             /**< IPX addr info (or internal)  */
  BYTE IPXImmediateAddress[6]; /**< IPX addr info (or internal)  */
  char IP[ADDR_SIZE];          /**< IP address as string */
  char IPh_addr[4];            /**< IP haddr (byte representation) */
  SINT32 portOffset;           /**< port offset (applied to all listening ports) */
  SINT32 inetProtocol;         /**< inet protocol (UDP, TCP, IPX, MMF, PIPE, etc.) */
  SINT32 tineProtocol;         /**< tine protocol level (highest level accepted) */
} FecDataStruct;
#define FECDATASTRUCT_SIZE (FEC_NAME_SIZE+16+16+4+3*4)
#define FDSR4_TAG "FDSr4"

typedef struct
{
  char os[12];
  char desc[64];
  char loc[32];
  char hdw[32];
  char sub[16];
  char resp[32];
} RPCFecExStruct;

typedef struct /* new */
{
  char os[FEC_OS_SIZE];
  char desc[FEC_DESC_SIZE];
  char loc[FEC_LOCATION_SIZE];
  char ver[FEC_VERSION_SIZE];
  char hdw[FEC_HDW_SIZE];
  char resp[FEC_RESP_SIZE];
} FecInfoStruct;

/**
 * \internal
 * \brief structure which contains the FEC address information
 *
 * The TINE library uses this structure for internal caches.
 *
 * \note struct is passed over the net (96 bytes)
 *
 */
typedef struct
{
  char fecName[FEC_NAME_SIZE]; /**< FEC name */
  FecNetAdr netAdr;            /**< network addr info */
  FecLclAdr lclAdr;            /**< local addr info (pipe, mmf) */
  char strAdr[ADDR_SIZE_IPV6]; /**< addr info as string */
  SINT32 portOffset;           /**< port offset */
  SINT32 inetProtocol;         /**< inet protocol type */
  SINT32 tineProtocol;         /**< tine protocol level */
  SINT32 addrFlags;            /**< optional address information */
} FecAddrStruct;
#define FECADDRSTRUCT_SIZE (FEC_NAME_SIZE+16+8+40+4*4) /* 96 bytes */
#define FASR5_TAG "FASr5"

typedef struct /* used for holding data temporarily during initialization */
{
  char fecName[FEC_NAME_SIZE];
  char fecOs[FEC_OS_SIZE];
  char fecContext[CONTEXT_NAME_SIZE];
  char fecSubsystem[SUBSYSTEM_NAME_SIZE];
  char fecDescription[FEC_DESC_SIZE];
  char fecLocation[FEC_LOCATION_SIZE];
  char fecHardware[FEC_HDW_SIZE];
  char responsible[FEC_RESP_SIZE];
  char expNameTarget[EXPORT_NAME_SIZE];
  char eqmNameTarget[EQM_NAME_SIZE];
  char historyHome[HISTORY_HOME_SIZE];
  char master[EXPORT_NAME_SIZE];
  char slaveMaster[EXPORT_NAME_SIZE];
  int  fecPort;
  int  fecMtu;
  int  fecBurstLimit;
  int  fecCycleDelay;
  int  fecWorkAreaSize;
} FIDS;

typedef struct
{
  int  devnum;
  int  devmask;
  int  devoffline;
  float zpos;
  char devname[DEVICE_NAME_SIZE];
  char devdesc[DEVICE_DESC_SIZE];
  char devloc[DEVICE_LOCATION_SIZE];
  char devrdr[DEVICE_REDIR_SIZE];
  char devregion[TAG_NAME_SIZE];
  char prplst[32];
} DEVIS;

typedef struct
{
  char FecName[FEC_NAME_SIZE];
  UINT32 numCalls;
  UINT32 localtime;
  UINT32 starttime;
  UINT32 systemPollingRate;
  short  numBkgTsks;
  short  numTotalContracts;
  short  numTotalClients;
  short  numTargetContracts;
  short  numTargetClients;
  short  numConnections;
  UINT32 numConnectionTimeouts;
  UINT32 numConnectionArrivals;
  UINT32 numUDPpkts;
  UINT32 numTCPpkts;
  UINT32 numIPXpkts;
  UINT32 numSPXpkts;
} ActivityQueryStruct;
#define ACTIVITYQUERYSTRUCT_SIZE (FEC_NAME_SIZE+4+4*3+2*6+4*6)
#define AQS_TAG "AQS"

typedef struct
{
  char desc[64];
  BYTE dataFmt;
  BYTE dataSiz;
  BYTE data[64];
  UINT32 timeEntered;
  UINT32 codeNow;
  UINT32 codePrv;
  UINT32 codeNxt;
} ServerStateStruct;
#define SERVERSTATESTRUCT_SIZE (64+2+64+4*4)

typedef struct
{
  UINT32 AveBusyTime;
  UINT32 CycleCounts;
  UINT32 MaxCycleCounts;
  UINT32 SingleLinkCount;
  UINT32 ClientMisses;
  UINT32 ClientReconnects;
  UINT32 ClientRetries;
  UINT32 ContractMisses;
  UINT32 ContractDelays;
  UINT32 BurstLimitReachedCount;
  UINT32 DataTimeStampOffset;
} ServerStatsStruct;
#define SERVERSTATSSTRUCT_SIZE (11*4)
typedef struct
{
  UINT32 srvWorkArea; /* srvWorkAreaSize */
  UINT32 cycDeadBand; /* gSystemCycleDeadband */
  UINT32 minPollInterval; /* MinPollingRate */
  UINT32 conTblCapacity; /* ContractListCapacity */
  UINT32 conRenewalLen; /* gSubRenewalLength */
  UINT32 clnTblCapacity; /* ClientListCapacity */
  UINT32 ackOnChange; /* RequireAcknowledgments */
  UINT32 srvBurstLimit; /* gBurstLimit */
  UINT32 srvBurstCycDly; /* gCycleDelay */
  UINT32 srvPacketMTU; /* gPacketMtu */
  UINT32 tcpMaxMsgSize; /* gMaxTcpMsgSize */
  UINT32 tcpTblCapacity; /* tcpConTblSize */
  UINT32 srvSendBuffers; /* max_srvrcvbuf */
  UINT32 srvRecvBuffers; /* max_srvsndbuf */
  UINT32 srvLazySched; /* gSchedulePropertyLazy */
  UINT32 srvBkgTskRentr; /* gAllowBkgTaskReentrancy */
  UINT32 srvCycInSepThrd; /* gRunServerCycleInSeparateThread */
  UINT32 tcpTraInSepThrd; /* gRunTransportInSeparateThread */
  UINT32 clnWorkArea; /* clnWorkAreaSize */
  UINT32 lnkTblCapacity; /* gConTblCapacity */
  UINT32 clnBurstLimit; /* gClientBurstLimit */
  UINT32 clnSendBuffers; /* max_clnrcvbuf */
  UINT32 clnRecvBuffers; /* max_clnsndbuf */
  UINT32 clnRecvQueue; /* gClnRecvQueueDepth */
  UINT32 useLoopbackAddr; /* gUseLoopback */
  UINT32 useWdogLinks; /* gAutoLinkWatchdogs */
  UINT32 allowCommonLnks; /* allowCommonLinks */
  UINT32 retryOnTimeout; /* gAlwaysRetry */
  UINT32 srvSchedIntvl; /*gSchedulerInterval; */
  UINT32 srvRetrdRemovl; /* gRetardSingleContractRemoval */
  UINT32 srvAllowRemote; /* gAllowRemoteExit, etc. */
  UINT32 reserved4;
} ServerSettingsStruct;
#define SERVERSETTINGSSTRUCT_SIZE (32*4)
#define SRVSETQS_TAG "SrvSetQy"
typedef struct
{
  char user[USERNAME_SIZE];
  char addr[ADDR_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  double timestamp;
} WriteAccessQuery;
#define WRITEACCESS_QUERY_SIZE sizeof(WriteAccessQuery)
typedef struct
{
  char user[USERNAME_SIZE];
  char addr[ADDR_SIZE_IPV6];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  double timestamp;
  int port;
  char eqm[EQM_NAME_SIZE];
} WriteAccessInfo;
#define WRITEACCESS_STRUCT_SIZE sizeof(WriteAccessInfo)
#define WRITEACCESS_TBL_SIZE 100
extern WriteAccessInfo wrAccTbl[WRITEACCESS_TBL_SIZE];
extern int wrAccTblPtr;

/**
 * \internal
 * \brief Determines the default value of MaxNumClients
 *
 * The places an upper limit on the number of simultaneous clients a server is willing
 * to manage.  If it is known that a server needs to service more clients than this
 * default, then the value of MaxNumClients should be set accordingly at initialization time.
 */
#ifndef CLIENTLIST_CAPACITY
# define CLIENTLIST_CAPACITY     100
#endif

/**
 * \internal
 * \brief Determines the default value of MaxNumContracts
 *
 * The places an upper limit on the number of simultaneous contracts a server is willing
 * to manage.  If it is known that a server needs to service more contracts than this
 * default, then the value of MaxNumContracts should be set accordingly at initialization time.
 */
#ifndef CONTRACTLIST_CAPACITY
# define CONTRACTLIST_CAPACITY     1000
#endif

#define CTR_RENEWAL     60
#define CTR_REMINDER    5
#define CTR_ACK_REQUEST (2*CTR_REMINDER)
#define CTR_MAXIMUM     32767
#define HBT_NONE        0
#define HBT_MINIMUM     1
#define HBT_REPLY_PENDING 3
#define HBT_MAXIMUM     60
#define HBT_SUB_PREPENDING 0x3ffe
#define HBT_SUB_PENDING 0x3fff
#define HBT_ACK_PENDING 0x7fff

/**
 * \internal
 * \brief The payload size of an IPX Datagram.
 *
 */
#define IPX_DATA_SIZE   512
#define IPX_BUFFER_SIZE 1466

/**
 * \internal
 * \brief The payload size of a UDP Datagram.
 *
 * \note TINE performs its own packet reassembly for incoming datagrams.  For many OSes
 * this is not necessary, but as packet reassembly cannot be assumed to be present or
 * might have an upper limit relating to the number of recv buffers, it is a fundemental
 * feature of TINE.
 */
#ifndef UDP_BUFFER_SIZE
# define UDP_BUFFER_SIZE 1472
#endif

#ifndef UDP_MAXBUFFER_SIZE
# define UDP_MAXBUFFER_SIZE 0x10000
#endif

#ifndef MMF_BUFFER_SIZE
# define MMF_BUFFER_SIZE 65000
#endif

#ifndef TCP_BUFFER_SIZE
# define TCP_BUFFER_SIZE 1024*16
#endif

#define REQ_BUFFER_SIZE 2000

/**
 * \internal
 * \brief The default value for one of srvWorkAreaSize, WorkAreaSize, and/or tmpWorkAreaSize.
 *
 * TINE uses a work area buffer for managing contracts calls.  On some (single-threaded) platforms
 * all work areas point to the same buffer.
 */
#ifndef WORK_SIZE
# define WORK_SIZE  0x10000 /* 64 KB */
#endif

/**
 * \internal
 * \brief The default value for one of clnWorkAreaSize.
 *
 * TINE uses a work area buffer for managing client calls.  On some (single-threaded) platforms
 * this points to the same area as the srvWorkAreaSize.
 */
#ifndef MIN_WORK_SIZE
# define MIN_WORK_SIZE 0x10000
#endif

/**
 * \internal
 * \brief The maximum payload size for incoming network globals.
 *
 */
#define GCAST_BUFFER_SIZE IPX_BUFFER_SIZE

/**
 * \internal
 * \brief The buffer size for a console command
 *
 */
#ifndef CMDSIZE
# define CMDSIZE 256
#endif

extern int nconsumers;
extern int ncontracts;
extern int StaleData;
extern int nExpiredContracts;

#define isWildcardContract(s) (strchr(s,'*') != NULL)
int EqmFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);
int EqmFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);

/**
 * \internal
 * \brief Contract input header for protocol level 6
 *
 * structure is passed over the net
 *
 */
typedef struct
{
  char   EqmProperty[PROPERTY_NAME_SIZE]; /**< requested property */
  char   EqmDeviceName[DEVICE_NAME_SIZE]; /**< requested device */
  char   EqmName[EQM_NAME_SIZE];          /**< requested local eqm name */
  UINT32 EqmSizeIn;                       /**< input data size */
  UINT32 EqmSizeOut;                      /**< output data size */
  BYTE   hEqmName;                        /**< handle to eqm name (unused) */
  BYTE   EqmAccess;                       /**< requested access (CA_READ etc) */
  BYTE   EqmFormatIn;                     /**< input data format */
  BYTE   EqmFormatOut;                    /**< output data format */
  char   strTagIn[TAG_NAME_SIZE];         /**< input data tag */
  char   strTagOut[TAG_NAME_SIZE];        /**< output data tag */
} CONTRACT;
#define CONTRACT_SIZE (PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SIZE) /* 180 bytes */

typedef struct
{
  CONTRACT contract;
  SINT32 pollingInterval;
  SINT32 nclients;
} ContractQueryStruct;
#define CONTRACT_QUERY_SIZE (CONTRACT_SIZE + 4 + 4)
#define CONQSR4_TAG "CTQSr4"

/* tineProtocol level 5  : TINE 3.0 - 3.31 */
/**
 * \internal
 * \brief legacy (release 3.xx) contract input header for protocol level 5.
 *
 * structure is passed over the net
 */
typedef struct
{
  char   EqmProperty[PROPERTY_NAME_SHORTSIZE];
  char   EqmDeviceName[DEVICE_NAME_SHORTSIZE];
  char   EqmName[EQM_NAME_SIZE];
  UINT32 EqmSizeIn;
  UINT32 EqmSizeOut;
  BYTE   hEqmName;
  BYTE   EqmAccess;
  BYTE   EqmFormatIn;
  BYTE   EqmFormatOut;
  char   strTagIn[TAG_NAME_SHORTSIZE];
  char   strTagOut[TAG_NAME_SHORTSIZE];
} CONTRACT3;
#define CONTRACT3_SIZE (PROPERTY_NAME_SHORTSIZE+DEVICE_NAME_SHORTSIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SHORTSIZE) /* 84 bytes */

typedef struct
{
  CONTRACT3 contract;
  SINT32 pollingInterval;
  SINT32 nclients;
} Contract5QueryStruct;
#define CONTRACT3_QUERY_SIZE (CONTRACT3_SIZE + 4 + 4)
#define CONQSR3_TAG "CONQS"

/**
 * \internal
 * \brief Legacy Client query struct
 *
 * struct is passed over the net (stock property 'CLIENTS')
 *
 */
typedef struct
{
  char userName[USERNAME_SIZE];
  IPXAddress IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;   /* IP address information */
  short ncontracts;  /* number of registrations */
  short tineProtocol;/* target client protocol level */
} CLNQS;

typedef struct
{
  CLNQS cln;
} ClientQueryStruct;
#define CLIENT_QUERY_SIZE (USERNAME_SIZE + 12 + 6 + 16 + 2 + 2)
#define CLNQSR3_TAG "CLNQS"

/**
* \internal
* \brief Modern Client query struct
*
* struct is passed over the net (stock property 'CLIENTS')
*
*/
typedef struct
{
  char userName[USERNAME_SIZE];
  char userAddr[ADDR_SIZE_IPV6];
  char userType[USERTYPE_SIZE];
  int pid;
  UINT16 port;
  short ncontracts;
  short tineProtocol;
  short inetProtocol;
} ClnInfoQueryStruct;
#define CLNQSR5_TAG "ClnInfoQS"

/**
 * \internal
 * \brief Client connection status struct used by TINE Engine.
 *
 */
typedef struct                /* local */
{
  short  id;                  /* contract ID as seen by consumer */
  short  mode;                /* SINGLE,DATACHANGE,TIMER,CANCEL */
  SINT32 pollingInterval;     /* requested polling interval */
  struct timeval lasttime;    /* time of most recent update */
  UINT32 starttime;           /* start of initial subscription */
  UINT16 counter;             /* counts remaining in subscription(s) */
  UINT16 Stale;               /* > 0 => send new data to consumer */
  UINT16 numblks;             /* total number of message blocks */
  UINT16 blknum;              /* block number of this message */
  short  statusCode;          /* interim completion */
  UINT16 mtu;                 /* maximal transport unit */
  UINT32 nextDataSize;        /* next outgoing data size in bytes */
  UINT32 nextDataPtr;         /* next outgoing data array pointer */
  UINT32 bytesRemaining;      /* number of bytes data remaining */
  UINT32 bytesTotal;          /* total number of bytes data */
  UINT32 misses;              /* data return after expiration */
  UINT32 reconnects;          /* client unexpectedly re-establishes link */
  int    isRetry;             /* the contract is being resubmitted by the client */
} ClnSts;

#define MAXLONGGRAMS 5
typedef struct
{
  SCKADR sckadr;              /* identify client by IP address */
  UINT16 subId;               /* supplied by the consumer upon registration */
  UINT16 blksin;              /* # blocks arrived from long datagram */
  UINT16 blkid;               /* block id of long datagram */
  UINT8  busy,
         reserved;
  UINT8  *data;
  UINT32 blktime;
  UINT32 SubsStarttime;
} ContractProgress;

typedef struct ClnLstStruct
{
  ClnHdr *client;
  ClnSts status;
  struct ClnLstStruct *next;
} ClnLst;

/**
 * \brief Client Information Structure used in GetCallerInformation.
 *
 */
typedef struct ClnInfoStruct
{
  short  accessMode; /**< this client's requested access mode: CM_SINGLE, CM_DATACHANGE, CM_TIMER, etc. */
  UINT16 reserved1;  /**< reserved field */
  SINT32 pollingRate;/**< this client's requested polling rate */
  UINT32 starttime;  /**< this client's start of initial subscription */
  UINT16 counter;    /**< this client's counts remaining in subscription(s) */
  UINT16 reserved2;  /**< reserved field */
  char userName[USERNAME_SIZE]; /**< this client's user name */
  SCKADR addr;       /**< this client's ip address */
  short ncontracts;  /**< this client's total number of contracts */
  short tineProtocol;/**< this client's tine protocol level */
  short inetProtocol;/**< this client's transport protocol: UDP, TCP, STREAM , etc.  */
  UINT16 reserved3;  /**< reserved field */
} ClnInfo;

#define CONTRACT_ACTIVE    0x00
#define CONTRACT_EXPIRED   0x01
#define CONTRACT_AVAILABLE 0x02

/**
 * \internal
 * \brief Contract List Table (local)
 *
 */
typedef struct
{
  ClnLst   *clnLst;             /**< assciated clnLst list */
  short    nclients;            /**< number of consumers for this contract */
  short    contractID;          /**< local contract ID */
  CONTRACT contract;            /**< see above */
  char     *extStringSpace;     /**< extended string buffer */
  UINT32   extStringSize;       /**< extended string size */
  char     *lngDeviceName;      /**< long device name buffer */
  char     *lngPropertyName;    /**< long property name buffer */
  SINT32   pollingInterval;     /**< in mseconds, smallest number wins ... */
  struct   timeval lasttime;    /**< most recent time the data polled */
  struct   timeval lastmarked;  /**< last time the data were marked STALE */
  struct   timeval starttime;   /**< start time of the current call */
  UINT32   dtimestamp;          /**< data timestamp */
  int      dtimestampUSEC;      /**< data timestamp (usec fraction) */
  int      datastamp;           /**< user data stamp */
  int      sysstamp;            /**< system data stamp */
  void     *prpDataRef;         /**< assignable reference pointer (used for e.g. persistence with 'not_ready') */
  BYTE     *dataOut;            /**< data to send  */
  BYTE     *dataIn;             /**< data received */
  BYTE     *dataBuf;            /**< data working area */
  BYTE     *dataOutArr;         /**< data working array for e.g. CF_STRING output */
  BYTE     **dataPtr;           /**< data pointer array for e.g. CF_STRING input */
  DTYPE    *dout;               /**< prepared output DTYPE */
  DTYPE    *din;                /**< prepared input DTYPE */
  UINT32   dataOutCapacity;     /**< size of dataOut allocation */
  UINT32   dataSizeOut;         /**< output data size < contract.dataSize */
  int      isStkCall;           /**< is a stock-property call ? */
  int      runInSeparateThread; /**< call contract in its own thread */
  int      callPending;         /**< eqm or stk call is pending (multi-threaded?) */
  int      deliveryPending;     /**< pending flag for doDelivery() vs doScheduler() */
  int      expired;             /**< expired flag for contract removal */
  int      compStatus;          /**< completion status */
  int      returnCode;          /**< contract completion return code */
  UINT32   delays;              /**< delayed scheduling due to deliveryPending */
  UINT32   misses;              /**< missed scheduling (i.e. > 2x polling rate) */
  char     compString[STATUS_SIZE];/**< completion string (redirection ?) */
  ExportListStruct *exp;        /**< associated export element */
  ExportPropertyListStruct *prp;/**< associated property element */
  int      listId;              /**< cross-reference to this entry */
  int      unbuffered;          /**< payload data are unbuffered : dataOut = dataBuf */
  int      accessFlags;         /**< additional access flags (CA_REPEAT, CA_STATIC) */
  int      signalPending;       /**< needs to call a signal handler */
  THRHANDLE threadId;           /**< thread id of thread if run in separate thread */
} ContractListStruct;

/**
 * \internal
 * \brief legacy (3.xx) producer header struct (reply to client)
 *
 * struct is passed over the net : tineProtocol 5 (out)
 *
 */
typedef struct
{
  UINT16 msgsize;               /**< message size in bytes */
  UINT16 subId;                 /**< supplied by the consumer upon registration */
  UINT16 retCode;               /**< completion or interim status */
  UINT16 numblks;               /**< total number of message blocks */
  UINT16 blknum;                /**< block number of this message */
  UINT16 mtu;                   /**< maximum transport unit */
  UINT16 timestampMSEC;         /**< msec fraction of data timestamp */
  BYTE   EqmFormat;             /**< format of returned data */
  BYTE   counter;               /**< tell consumer how much is left */
  UINT32 timestamp;             /**< data timestamp */
  UINT32 ClnStarttime;          /**< supplied by the consumer upon registration (@ byte 20)*/
} PrdrHdr3;
#define PHDR5_SIZE (7*2+2*1+2*4) /* 24 bytes */

#define CX_NULL      0x00
#define CX_RESPONSE  0x01
#define CX_STALE     0x02
#define CX_HEARTBEAT 0x04
#define CX_EVENT     0x08
#define CX_TIMER     0x10
#define CX_MESSAGE   0x20
#define CX_ERROR     0x40
#define CX_ATTACH    0x80
#define CX_RETRY    0x100

extern int gPacketMtu;
#define NPMTU_OLDWAY(np,m) ((np) == IPX ? IPX_BUFFER_SIZE :\
                     (np) == MMF ? MMF_BUFFER_SIZE :\
                     ((np) == TCP && (m) > UDP_BUFFER_SIZE) ? TCP_BUFFER_SIZE : UDP_BUFFER_SIZE)
#define NPMTU(np,m) ((np) == IPX ? IPX_BUFFER_SIZE :\
                     (np) == MMF ? MMF_BUFFER_SIZE :\
                     (np) == TCP ? TCP_BUFFER_SIZE : UDP_BUFFER_SIZE)
#define DGMTU(m,np,tp) ((np) != UDP || (tp) == 5 || (m) == 0 ? NPMTU(np,m) : (m))

/**
 * \internal
 * \brief producer header struct (reply to client)
 *
 * legacy : release 4.xx era
 * struct is passed over the net : tineProtocol 6 (out)
 *
 */
typedef struct
{
  UINT16 msgsize;        /**< message size in bytes */
  UINT16 subId;          /**< supplied by the consumer upon registration (con tbl id) */
  UINT16 retCode;        /**< return completion code from the eqm */
  UINT16 numblks;        /**< total number of message blocks */
  UINT16 blknum;         /**< block number of this message */
  UINT16 mtu;            /**< maximum transport unit */
  UINT16 EqmFormat;      /**< format of returned data */
  UINT16 counter;        /**< subscription counter (how much is left) */
  UINT16 tineProtocol;   /**< tine protocol level */
  UINT16 xferReason;     /**< transfer reason flag (one of the CX_ codes above) */
  UINT32 ClnStarttime;   /**< supplied by the consumer upon registration (@ byte 20)*/
  UINT16 stssize;        /**< size of error/status string space */
  UINT16 stscode;        /**< access status code */
  UINT32 timestamp;      /**< data timestamp */
  UINT32 timestampUSec;  /**< usec fraction of data timestamp */
  UINT32 userstamp;      /**< application setable data stamp */
  UINT32 sysstamp;       /**< systematic data stamp */
} PrdrHdr;

#define PHDR_SIZE (12*2+5*4) /* 44 bytes */

/**
 * \internal
 * \brief relevant offsets in the PrdrHdr (release 4.xx era)
 */
#define PHDR_STTM_OFFSET 20
#define PHDR_PROT_OFFSET 16
#define PHDR_STSSIZ_OFFSET 24
#define PHDR_SUBID_OFFSET 2
#define PHDR_CC_OFFSET 4
#define PHDR_FMT_OFFSET 12
#define PDHR_SYST_OFFSET 40
#define PDHR_URST_OFFSET 36

 /**
 * \internal
 * \brief subscription response header struct (reply to client)
 *
 * modern : release 5.xx era
 * msgsize and mtu are now UINT32
 * separate fields for returned data size in bytes and size in elements
 * struct is passed over the net : tineProtocol 7 (out)
 */
typedef struct
{
  UINT32 msgSize;        /**< this message size in bytes */
  UINT16 subId;          /**< supplied by the consumer upon registration (con tbl id) */
  UINT16 retCode;        /**< return completion code from the eqm */
  UINT16 numblks;        /**< total number of message blocks */
  UINT16 blknum;         /**< block number of this message */
  UINT32 mtu;            /**< maximum transport unit */
  UINT16 dataFormat;     /**< format of contract data */
  UINT16 counter;        /**< subscription counter (how much is left) */
  UINT16 tineProtocol;   /**< tine protocol level */
  UINT16 xferReason;     /**< transfer reason flag (one of the CX_ codes above) */
  UINT32 starttime;      /**< supplied by the consumer upon registration (@ byte 20)*/
  UINT16 stssize;        /**< size of error/status string space */
  UINT16 stscode;        /**< access status code */
  UINT32 timestamp;      /**< data timestamp (UTC wraps in 2038) */
  UINT32 timestampUSec;  /**< usec fraction of data timestamp */
  UINT32 userstamp;      /**< application setable data stamp */
  UINT32 sysstamp;       /**< systematic data stamp (wraps every 13 years @ 10 Hz continuous */
  UINT32 dataSize;       /**< size of contract data */
  UINT32 msgSizeInElements; /**< number of elements (of type format) returned in this message */
} SubRspHdr;
#define SUBRSPHDR_SIZE (9*4 + 10*2) /* 56 bytes */

/**
* \internal
* \brief relevant offsets in the SubRspHdr (release 5.xx era)
*/
#define SUBRSPHDR_STTM_OFFSET 24
#define SUBRSPHDR_PROT_OFFSET 20
#define SUBRSPHDR_STSSIZ_OFFSET 28
#define SUBRSPHDR_SUBID_OFFSET 4
#define SUBRSPHDR_CC_OFFSET 6
#define SUBRSPHDR_FMT_OFFSET 16
#define SUBRSPHDR_SYST_OFFSET 44
#define SUBRSPHDR_URST_OFFSET 40

/**
 * \internal
 * \brief Contract header struct
 *
 * struct is passed over the net : tineProtocol 6 (in)
 * TINE 4.xx era
 */
typedef struct subscription
{
  UINT16    msgsize;        /**< total message size in bytes */
  UINT16    extsize;        /**< extended string space */
  UINT16    mtu;            /**< maximum transport unit for incoming data */
  UINT16    numblks;        /**< total number of message blocks */
  UINT16    blknum;         /**< block number of this message */
  UINT16    id;             /**< block id */
  short     mode;           /**< requested transfer mode (e.g. CM_TIMER, CM_DATACHANGE) */
  short     blkid;          /**< (incremented) block id */
  UINT32    pollingInterval;/**< requested polling interval in msec */
  UINT32    starttime;      /**< UTC starttime of the contract (sent from client) */
  CONTRACT contract;        /**< contract request */
} SubInfoPkt;

#define SUBINFO_SIZE (8*2+2*4+CONTRACT_SIZE) /* 204 bytes */
#define SUB_MODE_LOCATION (PKTHDR_SIZE + 8)

/**
 * \internal
 * \brief Legacy (release 3.xx) contract header struct
 *
 * struct is passed over the net : tineProtocol 5 (in)
 * TINE 3.xx era
 */
typedef struct subscription3
{
  UINT16    msgsize;        /**< total message size in bytes */
  UINT16    mtu;            /**< maximum transport unit for incoming data */
  UINT8     numblks;        /**< total number of message blocks           */
  UINT8     blknum;         /**< block number of this message             */
  UINT16    id;             /**< block id */
  short     mode;           /**< requested transfer mode (e.g. CM_TIMER, CM_DATACHANGE) */
  short     blkid;          /**< (incremented) block id */
  UINT32    pollingInterval;/**< requested polling interval in msec */
  UINT32    starttime;      /**< UTC starttime of the contract (sent from client) */
  CONTRACT3 contract;       /**< contract request : 84 bytes */
} SubInfoPkt3;
#define SUBINFO3_SIZE (6*2+2*4+CONTRACT3_SIZE) /* 104 bytes */
#define SUB3_MODE_LOCATION (PKTHDR_SIZE + 8)

/**
* \internal
* \brief modern subscription request header struct
*
* struct is passed over the net : tineProtocol 7 (in)
* TINE 5.xx era
*/
typedef struct
{
  UINT32    msgsize;        /**< total message size in bytes */
  UINT32    extsize;        /**< extended string space (long device names) */
  UINT32    mtu;            /**< maximum transport unit for incoming data */
  UINT32    blkid;          /**< (incremented) block id */
  UINT16    numblks;        /**< total number of message blocks */
  UINT16    blknum;         /**< block number of this message */
  UINT16    id;             /**< subscription id  (connection table index @ client */
  UINT16    mode;           /**< requested transfer mode (e.g. CM_TIMER, CM_DATACHANGE) */
  UINT32    pollingInterval;/**< requested polling interval in msec */
  UINT32    starttime;      /**< UTC starttime of the contract (sent from client) */
  CONTRACT contract;        /**< contract request : 180 bytes */
} SubReqHdr;

#define SUBREQHDR_SIZE (6*4+4*2+CONTRACT_SIZE) /* 212 bytes */
#define SUB_MODE_LOCATION (PKTHDR_SIZE + 8)

#define SUBMTU(np,tp) (((np) == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE) - \
                       ((tp) == 5 ? SUBINFO3_SIZE : (tp) == 6 ? SUBINFO_SIZE : SUBREQHDR_SIZE) - \
                       (tp < 7 ? PKTHDR_SIZE : REQHDR_SIZE))

extern ContractListStruct **ContractList;
extern ClnHdr **ClnTbl;     /* registered consumer list */

extern SOCKET  udpSrvSck;   /**< UDP Producer Socket */
extern SOCKET  udpClnSck;   /**< UDP Consumer Socket */
extern SOCKET  mcastClnSck; /**< UDP MCast/BCast Consumer Socket */
extern SOCKET  netSrvSck;   /**< UDP MCast Net Services Socket */
extern SOCKET  glbClnSck;   /**< UDP Global listening Socket */
extern SOCKET  udpSyncSck;  /**< UDP trivial link Socket */
extern SOCKET  dcsEvnSck;   /**< UDP doocs event listening Socket */

void clientCycle(void);

typedef struct
{
  char prpName[PROPERTY_NAME_SIZE];
  UINT32 prpSize;          /* maximum supported array size (output) */
  UINT32 prpSizeIn;        /* maximum supported array size (input)  */
  short prpFormat;         /* format (output) */
  short prpFormatIn;       /* format (input) */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* output array type */
  char prpStructTag[TAG_NAME_SIZE]; /* output data structure tag */
  char prpDescription[64]; /* short description of property */
  int (*fcn)(ContractListStruct *cl,char *dataIn,char *dataOut);
  int runInSeparateThread; /* flag for potential slow pokes ... */
  int isFecProperty;       /* stock property refers to FEC and not the EQM */
} StockPropertyStruct;
#ifndef StockPropertyStructType
# define StockPropertyStructType StockPropertyStruct
#endif
TINE_EXPORT extern int gNrStockProperties;
TINE_EXPORT extern StockPropertyStructType gStockProperty[];

#define GetStockPropertyId(str) \
  getTableIndex(str,StockPropXRefTable,STOCKPROP_HASH_SIZE,(char *)gStockProperty,gNrStockProperties,OFFSETIN(StockPropertyStruct,prpName),sizeof(StockPropertyStruct))

#define GATE_NONE    0x00
#define GATE_BOOLEAN 0x01
#define GATE_FILTER  0x02
#define GATE_ARRAY   0x04

typedef struct
{
  char modName[16];
  int  (*fcn)(ContractListStruct *cl,char *dataIn,char *dataOut);
  int  runInSeparateThread; /* flag for potential slow pokes ... */
  int  isEmbedded; /* is meta tag embedded in the property (else 'endswith') */
} StockModifierStruct;
#ifndef StockModifierStructType
# define StockModifierStructType StockModifierStruct
#endif
extern int NrStockModifiers;
extern StockModifierStructType StockModifier[];
extern int gStockResourceBusy;
short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty,int localOnly);
short IsStaticProperty(char *Property);
int getMetaFormatAndSize(char *meta,short *fmt,int *siz);
int usesGateFunction(char *meta);
int isPropertyRedirected(ExportListStruct *el,ExportPropertyListStruct *pls,char *devName);
int isRegisteredMetaProperty(ExportListStruct *el, char *eqm, char *mprp, char *dev);
char *getContractKey(char *key,int clid);
int marshallStruct(DTYPE* d, int dir);

#ifndef MESSAGE_LOG_DEPTH
#define MESSAGE_LOG_DEPTH 500
#endif
typedef struct
{
  double ts;
  int status;
  char msg[128];
} MsgType;
extern MsgType MsgTbl[MESSAGE_LOG_DEPTH];
extern int gMsgTblPtr;
int msglog(int status,char *text,...);
TINE_EXPORT int GetMsgTable(MsgType *msgtbl,int *nummsg);
TINE_EXPORT void ClearMsgTable(void);
TINE_EXPORT void dumpMsgTableToFile(void);
TINE_EXPORT int GetMessageList(char *lstbuf,int bufsize);

#define MAX_ALIAS_ENTRIES 1024
#ifndef ALIAS_HASH_SIZE
# define ALIAS_HASH_SIZE 103
#endif
#define ALIAS_HASH_PAYOFF (ALIAS_HASH_SIZE/10)
typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char alias[DEVICE_NAME_SIZE];
} ALIAS_TABLE;
extern ALIAS_TABLE *gAliasTable;
extern int nAliasTableEntries;
extern HashXRefStruct **AliasXRefTable;

int makeAliasXRefTable(void);
int GetAliasTable(char *eqm);
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen);
#define getAliasTableIndex(str) \
   getTableIndex(str,AliasXRefTable,ALIAS_HASH_SIZE,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,alias),sizeof(ALIAS_TABLE))
#define getAliasTableIndexFromName(str) \
   getTableIndex(str,NULL,0,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,name),sizeof(ALIAS_TABLE))

typedef struct staTrgStruct
{
  char ctx[CONTEXT_NAME_SIZE];
  char key[GLOBALS_KEY_SIZE];
  NAME64 thisState;
  NAME64 prevState;
  struct staTrgStruct *nxt;
} StaTrg;
extern StaTrg *staTrgLst;
int SystemStartStateTrigger(const char *context,const char *stateKey);

typedef struct cycTrgStruct
{
  CYCBFCNP fcn;
  char eqm[EQM_NAME_SIZE];
  char prpLst[256];
  void *ref;
  struct cycTrgStruct *nxt;
} CycTrg;
extern CycTrg *gCycTrgLst;
extern SINT32 gCycleNumber;
extern SINT32 gCycleOffset;
extern int useCycleTrigger;

/**
 * \fn TINE_EXPORT void SetUseCycleTrigger(int value)
 * \brief Establishes whether a server listens for a CycleNumber trigger from a CYCLER.
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number. If for some reason this is not desired, this behavior can be
 * determined via this call.
 *
 * \param value (boolean) determines whether the server will listen for CycleNumber multicasts from
 * a CYCLER (default: TRUE).
 *
 * \see GetUseCycleTrigger
 */
TINE_EXPORT void SetUseCycleTrigger(int value);

/**
 * \fn TINE_EXPORT int GetUseCycleTrigger(void)
 * \brief Returns whether a server listens for a CycleNumber trigger from a CYCLER.
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number.
 *
 * \return whether the server will listen for CycleNumber multicasts from
 * a CYCLER (default: TRUE).
 *
 * \see SetUseCycleTrigger
 */
TINE_EXPORT int GetUseCycleTrigger(void);

/**
 * \fn TINE_EXPORT void SetSystemStampOffset(int cycleOffset)
 * \brief Establishes a system cycle offset
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the cycle
 * tag needs to be offset by some value, then this routine may be used to establish such an offset.
 *
 * \param cycleOffset is the desired cycle offset (counts) to be applied to the incoming cycle number
 * from the registered CYCLER. (Default = 0).
 *
 * \see GetSystemStampOffset
 */
TINE_EXPORT void SetSystemStampOffset(int cycleOffset);

/**
 * \fn TINE_EXPORT int GetSystemStampOffset(void)
 * \brief Returns the registered system cycle offset
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the cycle
 * tag needs to be offset by some value, then this routine may be used to obtain the currently registered
 * cycle offset.
 *
 * \return The currently registered system cycle offset (default = 0).
 *
 * \see SetSystemStampOffset
 */
TINE_EXPORT int GetSystemStampOffset(void);

/**
 * \fn TINE_EXPORT void SetSystemStampDelay(int cycleDelay)
 * \brief Establishes the system cycle delay
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the application of
 * the cycle tag needs to be delayed by some value, then this routine may be used to establish such a
 * cycle delay value (in milliseconds).
 *
 * \param cycleDelay is the desired cycle delay (milliseconds), which will must elapse before the incoming cycle number
 * from the registered CYCLER is to be applied to all readback data. (default = 0).
 *
 * \see GetSystemStampDelay
 */
TINE_EXPORT void SetSystemStampDelay(int cycleDelay);

/**
 * \fn TINE_EXPORT int GetSystemStampDelay(void)
 * \brief Returns the registered system cycle delay
 *
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the application of
 * the cycle tag needs to be delayed by some value, then this routine may be used to obtain the currently registered
 * cycle delay value (in milliseconds).
 *
 * \return The currently registered system cycle delay (default = 0 msec).
 *
 * \see SetSystemStampDelay
 */
TINE_EXPORT int GetSystemStampDelay(void);

/* RegisterCycleTriggerFunction() */
/**
 * \fn TINE_EXPORT int RegisterCycleTriggerFunction(CYCBFCNP fcn,char *eqm,char *prpLst,void *reference)
 * \brief Registers a cycle trigger callback dispatch function
 *
 * If a CYCLER is running in a server's context, then the server will receive 'Cycle Number' events
 * scheduled by the designated CYCLER server.
 * The cycle number will make use of the 'System Data Stamp' to tag all data sets obtained from the
 * server.  A server can also register a trigger function dispatch routine (or routines) to be called when a 'Cycle
 * Number' event occurs.  The dispatch routines will be called prior to setting the 'System Stamp' to the new
 * Cycle Number, which will be set following the execution of all dispatch routines.  Optionally, the server can
 * provide a property (or list of properties) to be scheduled following the dispatch execution.  This will ensure
 * that such properties will be called immediately following dispatch execution AND contain the most recent Cycle Number
 * as the 'System Data Stamp'.
 *
 * \param fcn is a reference to the cycle trigger dispatch routine to be called following the reception of a new
 * Cycle Number.  This must have the prototype: void (*fcn)(int cycleNumber,int cycleStatus,void *reference).
 * Thus, the dispatch routine will be called with the current cycle number, the cycle status (possibly 'link_timeout' if
 * no cycle is received within the assigned globals heartbeat (see SetGlobalsHeartbeat)), and an optional 'reference',
 * which is a void pointer of the caller's choosing (and will be returned to the caller in the dispatch routine).
 * \param eqm the local equipment module name of the desired central server (e.g. "BPMEQM")
 * \param prpLst is the property or properties which are to be 'scheduled' following the execution of the dispatch
 * routine.  If more than a single property is to be scheduled, this should be a string containing a comma separated
 * list.
 * \param reference is a caller supplied void pointer which will be returned to the called in the dispatch
 * routine.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_cycletrigger.c
 *
 * \see UnregisterCycleTriggerFunction
 */
TINE_EXPORT int RegisterCycleTriggerFunction(CYCBFCNP fcn,char *eqm,char *prpLst,void *reference);

/**
 * \fn TINE_EXPORT int UnregisterCycleTriggerFunction(CYCBFCNP fcn,void *reference)
 * \brief Unregisters a previously registered cycle trigger callback dispatch function
 *
 * If a cycle trigger event dispatch routine is no longer required, it can be unregistered using this
 * routine.
 *
 * \param fcn is a reference to the previously registered cycle trigger dispatch routine which is to
 * be removed.
 * \param reference is a caller supplied void pointer which will be returned to the called in the dispatch
 * routine.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterCycleTriggerFunction
 */
TINE_EXPORT int UnregisterCycleTriggerFunction(CYCBFCNP fcn,void *reference);

/**
 * \fn TINE_EXPORT int RegisterStateChangeCallback(SCCBFCNP fcn,const char *eqm,const char *stateKey, void *reference)
 * \brief Registers a state change callback dispatch function
 *
 * If STATE and GLOBALS servers are running, a server can receive state change notifications and
 * react accordingly by calling this
 * routine and passing a callback dispatch routine.  The callback routine will be called only upon change of state.
 * In the event of any i/o errors which prevent the server from receiving the state information, the state will be
 * changed automatically to 'unavailable' following 5 consecutive readback errors.
 *
 * \param fcn is a reference to the state change dispatch routine to be called following a change of state
 * This must have the prototype: void (*fcn)(const char *previousState,const char *currentState,void *reference).
 * If a NULL is passed for this parameter then any callback routine will be de-registered.
 * \param eqm the local equipment module name of the desired central server (e.g. "BPMEQM")
 * \param stateKey is an optional specification of the desired state keyword from the GLOBALS server.
 * If a NULL is passed, then the default of /\<context\>/GLOBALS/DeclaredState will be used.
 * \param reference is a caller supplied void pointer which will be returned to the called in the dispatch
 * routine. A NULL can be passed if no reference is required.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_statetrigger.c
 *
 */
TINE_EXPORT int RegisterStateChangeCallback(SCCBFCNP fcn,const char *eqm,const char *stateKey, void *reference);

TINE_EXPORT int SystemGetSiteTimestamp(double *timeStamp);
TINE_EXPORT int SystemStartTimeSynchronization(void);
TINE_EXPORT UINT32 SystemGetMemory(void);

#define CDI_LNK_TBL 0x1000000
#define MAKE_CDI_LINKID(id) (id < 0 ? id : ((id) | CDI_LNK_TBL))
#define GET_CDI_LINKID(id) ((id) & CDI_LNK_TBL ? (id) & ~(CDI_LNK_TBL) : -1)
extern TINE_EXPORT void (*cdiSystemInitHook)(void);
extern TINE_EXPORT int (*cdiCloseLinkHook)(int);
extern TINE_EXPORT char *(*getCdiLinkNameHook)(char *,char *);
extern TINE_EXPORT int (*cdiExecLinkHook)(char *,char *,DTYPE *,DTYPE *,short,UINT16);
extern TINE_EXPORT int (*cdiAttachLinkHook)(char *,char *,DTYPE *,DTYPE *,short,int,TCBFCNP,int,UINT32);
extern TINE_EXPORT void (*cdiSystemFreeHook)(void);

extern SYSTSKP _preSysIniFcn;
extern SYSTSKP _postSysIniFcn;
extern SYSTSKP _postSysCycFcn;
extern int gIsCyclePollingMode;
extern TINE_EXPORT int gRequireAcknowledgments;
#define RequireAcknowledgments gRequireAcknowledgments
extern int AveBusyTime;
extern struct timeval thisCycleTime;
extern BYTE *BrdcstBuffer;
extern UINT32 TCPcumcnt,TCPinitcnt;
extern time_t gSystemStartupTime;
extern int hConn, hControl;
extern int ipLoaded,ipxLoaded,numErr;
extern char gUserName[20];      /* login name of PC user */
extern TINE_EXPORT char gFecName[FEC_NAME_SIZE+4];
extern BYTE *clientbuf;
extern BYTE *WorkArea;
extern BYTE *tmpWorkArea;
extern UINT32 WorkAreaSize;
extern UINT32 tmpWorkAreaSize;
extern BYTE *mcstWorkArea;

extern char gFecAddrBuffer[64];
extern char gMyipAddrBuffer[64];
extern int gListenForMCasts;
extern int foregroundTTY;
extern UINT32 CMDcount;

extern ConTblEntry **conTbl;
extern GrpTblEntry *grpTbl;
extern int ReturnToClientCycle;
extern int ReturnToServerCycle;
extern int nConnectionTableEntries;
extern SOCKET tcpSrvSck,tcpStrmSck,dbgSrvSck;
extern int CheckDeniedList;
extern BYTE spxSrvSckB[2];
extern const BYTE ipxClnSckB[2];
extern BYTE ipxSrvSckB[2];
extern const BYTE ipxMcastSckB[2];
extern char PATH[];
extern FecAddrStruct *FecTbl, *NameServer;
extern ExpDataStruct *SrvTbl;
extern FecInfoStruct gFecInfo;
extern int numNameServers,curNameServer;
extern int numSrvTblEntries,numFecTblEntries,FecTblSize;
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount,BRDcount;
extern UINT32 nConnectionTimeouts,nArrivals;
extern UINT32 nPartialTransfers;
extern short gNumEqmBkgTasks;
extern time_t *tnull;
extern TINE_EXPORT char gDeviceContext[CONTEXT_NAME_SIZE+4];
extern TINE_EXPORT char gDeviceSubSystem[SUBSYSTEM_NAME_SIZE+4];
extern UINT32 gContractMisses;
extern UINT32 gContractDelays;
extern UINT32 gClientMisses;
extern UINT32 gClientReconnects;
extern UINT32 gClientRetries;
extern UINT32 gBurstLimitReachedCount;
extern UINT32 gSingleLinkCount;
extern int gCycleCounts, gMaxCycleCounts;
extern int gLastMinuteInitializationDone;
extern int max_clnrcvbuf;
extern int max_srvrcvbuf;
extern int max_clnsndbuf;
extern int max_srvsndbuf;
extern int max_udpsndbuf;
extern TINE_EXPORT int gRetardSingleContractRemoval;
extern TINE_EXPORT int ServerExitCondition;
extern TINE_EXPORT int ServerInitialized;
extern TINE_EXPORT int ServerWaiting;
extern TINE_EXPORT int gIgnoreMetaPropertyQueries;
extern int acquireDeviceNamesFromProperties;
extern int canSendGlobals;
extern char *ltzname[2];
extern char tExecLinkReturnedCompletion[STATUS_SIZE+4];
extern int tExecLinkReturnedDataSize;
extern int gTineProtocol;
extern int gGlobalsProtocol;
TINE_EXPORT void SetSendLegacyGlobals(int value);
TINE_EXPORT int GetGlobalsProtocol(void);
extern int gEqmFcnContractTableId;
extern int gEnforceMcaAccess;
extern int gClientCycleIdle;
extern int gServerCycleIdle;
extern int gIgnoreCentralServices;
extern int gInitIPServicesDone;
extern int gInitIPClientDone;
extern int gInitClientDone;
extern int isUsingCycleTrigger;
extern int gAutoLinkErrorAlarms;
extern int gUseSynchronousNotification;
extern int hasLogFileAlarm;
extern int disallowSingleDelivery;

extern TINE_EXPORT int gServerCycleState;
#define CYCLE_STATE_NONE 0
#define CYCLE_STATE_IDLE 1
#define CYCLE_STATE_INIT 2
#define CYCLE_STATE_INIT_FINAL 3
#define CYCLE_STATE_DELAY 4
#define CYCLE_STATE_ACCEPT 5
#define CYCLE_STATE_SCHED 6
#define CYCLE_STATE_EQM 7
#define CYCLE_STATE_ACQNAMES 8
#define CYCLE_STATE_CLEANUP 9
#define CYCLE_STATE_ALMWATCH 10
#define CYCLE_STATE_ALMPROC 11
#define CYCLE_STATE_HST_INIT 12
#define CYCLE_STATE_HST_MKFILES 13
#define CYCLE_STATE_HST_EQM 14
#define CYCLE_STATE_HST_WRITE 15
#define CYCLE_STATE_HST_RMVFILES 16
#define CYCLE_STATE_SNDEXP 17
#define CYCLE_STATE_CHK_LOCKS 18
#define CYCLE_STATE_CHK_GRP 19
#define CYCLE_STATE_CHK_FO 20
#define CYCLE_STATE_EXIT 21

#define isValidConTblEntry(i) (((i) >= 0) && ((i) < nConnectionTableEntries) && (conTbl[i] != NULL))

/**
 * \fn TINE_EXPORT void SetInterpretConsoleCommands(int value))
 * \brief Establishes the ability of the TINE kernel to parse input if a console is available.
 *
 * Console commands such as setting the debug level or dumping statistics can be given at
 * the application's console if it is running in the foreground if this setting is TRUE. This
 * routine has no effect if the system is not running on a UNIX or UNIX-like OS. WINDOWS
 * console applications are always able to parse and interpret any console input.
 *
 * \param value is the desired setting.
 */
TINE_EXPORT void SetInterpretConsoleCommands(int value);

/**
 * \fn TINE_EXPORT int GetInterpretConsoleCommands(void)
 * \brief Returns the setting for whether console commands will be interpreted or not.
 *
 * Console commands such as setting the debug level or dumping statistics can be given at
 * the application's console if it is running in the foreground if this setting is TRUE. This
 * routine has no effect if the system is not running on a UNIX or UNIX-like OS. WINDOWS
 * console applications are always able to parse and interpret any console input.
 *
 * \return the current setting.
 */
TINE_EXPORT int GetInterpretConsoleCommands(void);

/**
 * \fn TINE_EXPORT int GetServerCycleState(char *cycleStateString)
 * \brief Returns the current server cycle state
 *
 * This routine can be polled or otherwise called at any time to
 * determine what state the server cycle is in.
 *
 * \param cycleStateString is an optional string buffer which if
 * present will be filled with the current cycle state as a
 * representative sting.  The parameter passed must reference a
 * character buffer of at least 64 characters.
 *
 * \return a reference to the the Export List Structure of the given
 * equipment module or NULL if the equipment module name is not found in the registry.
 *
 * \see ExportListStruct for details
 */
TINE_EXPORT int GetServerCycleState(char *cycleStateString);

void toFecAddr(FecAddrStruct *fas,FecDataStruct *fds);
void toFecData(FecDataStruct *fds,FecAddrStruct *fas);
TINE_EXPORT void SetServerExitCondition(int value);
/**
 \internal if reattaching to multicast group is relevant
 */
void reAttachToServiceGroup(void);
int SystemStartCycleTrigger(void);
int makeValidPollingInterval(int pollingRate, int baseMode);
char *makeListFileName(char *eqm,int usesubdir,char *prefix,char *postfix);
void sendPrpSignal(ExportPropertyListStruct *prp,int signal,int conId,int sts);
void prepOutgoingData(BYTE *dst,BYTE *src,UINT32 dsiz,short fmt);
void prepIncomingData(BYTE *dst,BYTE *src,UINT32 dsiz,UINT32 rsiz,short fmt,short endianness,ConTblEntry *c);
int prepIncomingSubResponseHdr(int tineProtocol, SubRspHdr *phdr, BYTE *stream);
int prepOutgoingSubResponseHdr(ClnLst *cn,ContractListStruct *cls,UINT32 msgsize,BYTE *stream);
int prepIncomingResponseHdr(RspHdr *hdr, RspSpecs *spc, BYTE *stream, UINT32 totalSizeInBytes);
void prepOutgoingResponseHdr(ClnHdr *cln, RspHdr *hdr, BYTE *stream);
int prepOutgoingSubRequestHdr(int tineProtocol, ConTblEntry *c, char *eqm, int datasize, int subid, BYTE *pSubstream);
int prepIncomingSubRequestHdr(int tineProtocol, SubReqHdr *hdr, BYTE *stream, int *swallowError);
int prepOutgoingRequestHdr(int protocol, UINT32 totalSizeBytes, int segments, const char *usr, const char *typ, BYTE *hdr);
int prepIncomingRequestHdr(ReqHdr *hdr, BYTE *stream);

#define HDRRSP_SIZE(p) ((p) < MINIMUM_REQHDR_PROTOCOL ? sizeof(UINT16) : RSPHDR_SIZE)
#define SUBREQ_STEPSIZE(p) ((p) == 5 ? SUBINFO3_SIZE : (p) == 6 ? SUBINFO_SIZE : SUBREQHDR_SIZE)
#define SUBRSP_STEPSIZE(p) ((p) == 5 ? PHDR5_SIZE : (p) == 6 ? PHDR_SIZE : SUBRSPHDR_SIZE)
#define PMTU(m,np,tp)  (DGMTU(m,np,tp) - SUBRSP_STEPSIZE(tp) - HDRRSP_SIZE(tp))

NAME64 *GetEqpDeviceNames(char *eqm,char *prpname,ExportDeviceType **devs,int *listSize);
int IsIllegalProperty(char *eqmName,char *prpName,char *dev,ClnHdr *cln,short access,
                      ExportListStruct **expLstElem,ExportPropertyListStruct **prpLstElem,char *stsString);
int doContract(ContractListStruct *cl,int canUseSeparateThread);
void doGCast(struct timeval *thisTime);
int GetPrpId(char *eqm,char *prpName);
void devNamesInitialization(void);
void ExitEQPModules(void);
void checkTasklist(void);
int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen);
char *getFirstDevice(char *eqm,char *prpname);
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp);
int resetPropertyDeviceList(char *eqm,char *prp);
int prepSubRequest(int isENS);
int sendSubRequest(int fecIdx,SOCKET sck,int inetProtocol,BYTE *data,UINT32 sizeInBytes);
#define SCHED_INTERN 0
#define SCHED_RESPONSE 1
#define SCHED_EXTERN 2
int doScheduler(struct timeval *thisTime,int timeout,int source,ClnHdr *tgtcln,ContractListStruct *tgtcon);
int doDelivery(struct timeval *thisTIme,ClnHdr *tgtcln,ContractListStruct *tgtcon,int tcponly);
int sendToPeer(ClnHdr *c,BYTE *payload,UINT32 msgsize,int msgtype,int mark);
int getConsumerRequest(ClnHdr *cln,BYTE *buf,int packetsize);
int getConTblEntry(DTYPE *dout,DTYPE *din,char *devcontxt,char *devgrp,char *devname,char *property,short access,int reuseIndex);
int dumpConsumerList(void);
int dumpContractList(void);
int dumpMessageList(void);
int dumpEqmList(void);
int dumpEqmProperties(char *eqm);
int dumpEqmDevices(char *eqm);
int dumpMcaDeviceMembers(char *eqm);
int dumpEqmHistories(char *eqm);
int dumpEqmAlmWatch(char *eqm);
int dumpConnectionList(void);
int dumpRedirectionList(void);
void dumpRegisteredUsers(void);
int dumpConnectionGroups(void);
int dumpSemaphores(void);
int dumpFecAddr(void);
int dumpWcList(void);
void prepDataDump(char *buf,char *data,int siz,short fmt,int output);
UINT32 doProcessData(UINT32 bufsize,BYTE *buf,void *padr,int inetprot,char *ptag);
int initFecTable(void);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
void getGlobalUpdate(SCKADR *sckadr,BYTE *buf);
void getDcsEvnUpdate(SCKADR *sckadr,BYTE *buf);
void quitServer(void);
void hardDelay(int msec);
int isMemberLockSet(ExportListStruct *el,ClnHdr *cln);
void lockToExclusiveRead(ClnHdr *cln);
int removeClientFromContract(ClnLst *c,int tid);
int removeClientTableEntry(int tid);
int removeContractTableEntry(int tid);
int isExclusiveProperty(ExportListStruct *el,char *prp);
int isAllowedUser(char *eqm,char *usr);
int isDeniedUser(char *usr);
int isDeniedNet(SCKADR *adr);
int isAllowedPropertyAccess(const char *eqm,const char *prp,ClnHdr *cln);
int isAllowedDeviceAccess(const char *eqm,const char *dev,ClnHdr *cln);
ClnHdr *getNetworkClientFromList(void);
ClnHdr *LocateConsumerInList(SCKADR *ip,IpxAdr *nwAddr,void *hwnd);
TINE_EXPORT UINT32 ElfHash(const unsigned char *key);
void makeEqpDevicesHashTable(ExportListStruct *el);
int makeStockPropXRefTable(void);
short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate);
TINE_EXPORT short IsStockProperty(char *Property);
short IsFecProperty(char *Property);
int hasVolatileDataSize(int fmt,char *tag,char *ctx,char *srv);
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName,ExportPropertyListStruct **prp);
char *lookupPropertyId(ExportListStruct *el,short id);
int sndRegisteredExports(struct timeval *tv);
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp);
void ParseRedirector(char *str,PrpRedirBlk *prd);
ExportListStruct *getExportListItem(char *eqm);
ExportListStruct *getExportListItemFromId(int id);
int SystemStartGlobalSynchronization(void);
TINE_EXPORT int reformat(BYTE *dstData,int dstFmt,BYTE *srcData,int srcFmt,int num);
int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv);
TINE_EXPORT int GetServerInitialization(void);
char *getPrimaryContext(char *ctx,char *sub);
char *makeLinkKey(char *key,char *context,char *server,char *device,char *property);
char *getLinkKey(ConTblEntry *c);
void flushGlobalsTable(int freeMemory);
int hasFilesOfType(const char *path,const char *pattern);
int isTraceLink(int id);
void traceLinkOutput(ConTblEntry *c, char *tag, char *msg);
int IsInForeground(void);

int _recvGlobal(const char *keyword,DTYPE *dout,void *callback,UINT32 callbackID,void *reference);
int _attachLink(const char *devName,const char *devProp,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void *callback,int mode,UINT32 callbackID,void *reference,int reuseIndex);
void _closeLink(int linkId);
void NotifyClient(int i,int cc);
void NotifyGlobalClient(int i,int cc);
int connectTCP(ConTblEntry *c,int port,int inetProtocol);
int initIPServer(void);
int initClient(void);
int sendIpData(ClnHdr *cln,BYTE *data,UINT32 msgsize, int msgtype, int inetProtocol,int yield,int canFreeOnError);
int sendStreamData(ClnHdr *cln,BYTE *data,int msgsize);
int sendIPgcast(BYTE *buf,int size);
ClnHdr *getConsumerIPAddress(SOCKET sck,int inetProtocol);
int redirectLink(char *con,char *srv,char *prp,char *dev);
void dumpIPNets(void);
void AcceptIP(struct timeval *tv);
void CleanupIP(struct timeval *tv);
void closeIPServer(void);
int IsMemberIPNets(int num,SCKACL *lst,SCKADR *ipaddr);
int tUDPCall(int fecindex,BYTE *buf,UINT32 msgsize,int timeout);
int tExecLink(char *fec,char *eqm,char *dev,char *prp,DTYPE *dout,DTYPE *din,short access,short timout);
void checkClientNotification(void);
void checkSubscriptions(int isENS);
int isAbsolutePath(char *fname);
int readSelfTest(char *eqm,char *textbuf,int sizeInBytes);
int readFecLog(char *textbuf,int sizeInBytes);
int readFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int writeFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int readFecFile(char *fname,char *buf,int *sizeInBytes);
int writeFecFile(char *fname,char *buf,int sizeInBytes);
int scanFecLogFiles(char *path,char *subpath,void *namebuf,int fmt,int *numfiles,int numoffset);
int getFecLogFileLength(char *fname,int *mtime);
void applyFecLogPath(char *path,int pathlen);
void checkAccessLocks(time_t t);
void checkGroupMembers(time_t t);
void checkFailoverStatus(time_t t);
TINE_EXPORT void haltServer(void);
void freeServerMemory(void);
TINE_EXPORT void FreeEQPModules(void);
TINE_EXPORT int dbglog(char *text,...);
int isSpectrumArray(char *eqm,char *prpName);
int isChannelArray(char *eqm,char *prpName);
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken);
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen);
int isContiguous(char *eqm,char *devName,NAME64 *devList,int *start,int *len);
int isLocal(char *eqm,char *devName,char *prpName);
ExportPropertyListStruct *GetPropertyListStruct(char *eqm,char *prpName,char *devName);
ExportPropertyListStruct *GetPropertyListStructEx(char *eqm,char *prpName,char *devName,int flags);
int queryBitfieldFormat(char *srv,char *tag,NAME16I *bffq,int *bffqsiz);
int callEqmInSeparateThread(char *eqm,char *prp,int access);
extern int gClnRecvQueueDepth;
int callEqm(ContractListStruct *cl,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
int IsPropertyCalledInSeparateThread(ContractListStruct *cl);
void startFailoverMonitor(char *eqm);
void _cbXRef(int id, int cc);
void _cbMcaXRef(int id, int cc,void *ref);
int getCbXRefTblId(void);
void flushContractTable(void);
void sendContractSignal(const char *eqm,const char *dev,const char *prp,int signal,int sts);
ReLnkTblEntry *getReLnkLstItem(char *ctx,char *srv,char *dev,char *prp,int siz,int fmt);
ReLnkTblEntry *getReLnkTblEntry(ConTblEntry *c);
void appendReLnkLst(ConTblEntry *c,char *tag,int fmt,int siz,int sts);
ReLnkTblEntry *relinkLink(char *ctx,char *srv,char *prp,char *dev,char **pkey,DTYPE *dout);
int reAttachLink(int i,int isStructField);
void flushReLnkLst(void);
int isJumboDelivery(ClnHdr *cln);

int isBitfieldAccess(DTYPE *d);
int isBitfieldProperty(char *srv,char *tag,char *prp,char *fld);
UINT32 getBitfieldValue(int fmt,BYTE *dptr);
int splitFullName(const char *fullName,char *srvName,char *devName);
int GetIntValue(DTYPE *d,SINT32 *ival);
int isTaggedStructAccess(DTYPE *d);
int isStructProperty(char *tag,char *prp,char *fld);
void mapAndCallbackBitfield(ConTblEntry *c);
void mapAndCallbackStructField(ConTblEntry *c);
void mapAndCallbackData(ConTblEntry *c);
UINT32 getAvailableMemoryBytes(void);

TINE_EXPORT int tokenizeFullName(char *fullName,char **ctx,char **srv,char **dev);
TINE_EXPORT void ResetLastMinuteInitialization(void);
TINE_EXPORT void resetLastMinuteInitialization(void);
void fixFecRepository(void);

typedef void (*ISETFCN)(int);
typedef void (*DSETFCN)(double);

typedef struct
{
  char name[32];
  void *attr;
  ISETFCN fcn;
  int type;
} SystemAttribute;
extern SystemAttribute systemAttributes[];
extern int nSystemAttributes;
TINE_EXPORT int SetSystemAttribute(char *attribute,void *value,int format);
TINE_EXPORT int GetSystemAttribute(char *attribute,void *value, int format);
TINE_EXPORT int GetSystemAttributeList(NAME64I *lst, int *length);

#ifndef SMALL_TINE_LIB
extern DeviceRegionItem *_drlstbuf;
extern int _drlstbufsiz;
extern int _drlstsiz;
int URLQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut);
int historyRecordsAppend(ContractListStruct *cl,char *dataIn,char *dataOut);
int globalsRecordsQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int srvExit(ContractListStruct *cl,char *dataIn,char *dataOut);
int srvReset(ContractListStruct *cl,char *dataIn,char *dataOut);
int srvInit(ContractListStruct *cl,char *dataIn,char *dataOut);
int srvReex(ContractListStruct *cl,char *dataIn,char *dataOut);
int srvIdle(ContractListStruct *cl,char *dataIn,char *dataOut);
int setDebugLevel(ContractListStruct *cl,char *dataIn,char *dataOut);
int setLogCommands(ContractListStruct *cl,char *dataIn,char *dataOut);
int setLogDepth(ContractListStruct *cl,char *dataIn,char *dataOut);
int getAppDate(ContractListStruct *cl,char *dataIn,char *dataOut);
int getAppVersion(ContractListStruct *cl,char *dataIn,char *dataOut);
int getFecLogFileList(ContractListStruct *cl,char *dataIn,char *dataOut);
int getManifest(ContractListStruct *cl,char *dataIn,char *dataOut);
int getSrvManifestPath(ContractListStruct *cl,char *dataIn,char *dataOut);
int getSrvCmdLine(ContractListStruct *cl,char *dataIn,char *dataOut);
int getSrvCwd(ContractListStruct *cl,char *dataIn,char *dataOut);
int getSrvPid(ContractListStruct *cl,char *dataIn,char *dataOut);
int almFilterOuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int hstFilterQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int prepIPNetLst(CONTRACT *con,char *dataIn,int *siz,NAME64 **lst);
int addRegisteredIPNet(ContractListStruct *cl,char *dataIn,char *dataOut);
int rmvRegisteredIPNet(ContractListStruct *cl,char *dataIn,char *dataOut);
int getNumRegisteredIPNets(ContractListStruct *cl,char *dataIn,char *dataOut);
int getRegisteredIPNetList(ContractListStruct *cl,char *dataIn,char *dataOut);
int getNumRegisteredIPXNets(ContractListStruct *cl,char *dataIn,char *dataOut);
int getRegisteredIPXNetList(ContractListStruct *cl,char *dataIn,char *dataOut);
int nalarmWatchQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int alarmWatchQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int localAnnotatedCMTQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int localAnnotatedCMTSQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int metapropsQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int getSrvAddressParams(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceLocQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceRegionQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceOnlineQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceMaskQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceZPosQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int deviceDescQuery(ContractListStruct *cl,char *dataIn,char *dataOut);
int getEnsList(ContractListStruct *cl,char *dataIn,char *dataOut);
int getServerAliasList(ContractListStruct *cl,char *dataIn,char *dataOut);
int addAlias(ContractListStruct *cl,char *dataIn,char *dataOut);
int setAclFileInfo(char *eqm,int *usrFile,int *netFile);
int isAllowedSaveRestoreProperty(ExportListStruct *el,ExportPropertyListStruct *pl);
int adjustWorkAreaSize(UINT32 newSizeInBytes);
int joinEnsGroup(ExportListStruct *el);
int prepAdjustableArrayOutputPreCall(ContractListStruct *cl,DTYPE *dout);
int prepAdjustableArrayInputPreCall(ContractListStruct *cl,DTYPE *din);
int prepAdjustedArrayOutputPostCall(ContractListStruct *cl,DTYPE *dout);
int removePhantomClientFromContracts(ClnHdr *cln,char *tag,int needLock);
int isValidClient(ClnHdr *cln);
void appendReLnkCb(ConTblEntry *c);
int hasVolatileDataSize(int fmt,char *tag,char *ctx,char *srv);
int assertBufferCapacity(ConTblEntry *c,int numblks,int mtu,int isVolatile);
void lwdCloseLink(LnkWatchDog *lwd);
DTYPE *remapAdjustableInputData(DTYPE *din);
void clearDependentLinks(int linkId);
void removeDependentLink(int linkId);
void wcMembersCallback(int id,int cc,void *ref);
int assignWildcardLink(int lid);
#define IsMcaParentLink(c) ((c)->cbfcn == (TCBFCNP)_cbMcaXRef)
#endif

/**
 * \fn TINE_EXPORT ExportListStruct *GetExportListItem(char *eqm)
 * \brief Returns a reference to the Export List Structure of the given
 * equipment module.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return a reference to the the Export List Structure of the given
 * equipment module or NULL if the equipment module name is not found in the registry.
 *
 * \see ExportListStruct for details
 */
TINE_EXPORT ExportListStruct *GetExportListItem(char *eqm);

/* GetCallbackGroup() */
/**
 * \fn TINE_EXPORT GrpTblEntry *GetCallbackGroup(size_t id)
 * \brief Returns a reference to the callback Group Table Entry associated with the
 * identifier supplied
 *
 * \param id is the group callback id (effectively the reference pointer to the callback
 * function associated with the group.
 *
 * \return a reference to the group table entry (GrpTblEntry *) if the given id corresponds
 * to a registered callback group, or NULL if not found.
 *
 * \see GetGroupMemberList(), GrpTblEntry, GrpMember
 */
TINE_EXPORT GrpTblEntry *GetCallbackGroup(size_t id);

/* GetGroupMemberList() */
/**
 * \fn TINE_EXPORT GrpMember *GetGroupMemberList(GrpTblEntry *grp)
 * \brief Returns a linked list of the group members for the callback group supplied
 *
 * \param grp is a reference to the group table entry (GrpTblEntry *) whose member list
 * is desired.
 *
 * \return a reference to a linked list of the group members for the callback group supplied
 *
 * \see GetCallbackGroup(), GrpTblEntry, GrpMember
 */
TINE_EXPORT GrpMember *GetGroupMemberList(GrpTblEntry *grp);

/* FindServerOnNetwork() */
/**
 * \fn TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecAddrStruct *fec,ExpDataStruct *srv)
 * \brief Issues a multicast (or broadcast) to which configured TINE central servers respond
 *
 * Diskless (or otherwise 'in-a-box') servers which need to find central servers such as the
 * Equipment Name Server, Central Alarm Server, or Post-Mortem Server, etc. can issue this
 * call to ascertain the address.
 *
 * \param context (optional) the Context of the desired central server
 * \param eqmName the local equipment module name of the desired central server (e.g. "ENSEQM")
 * \param exportName the exported device server name of the desired central server. Note either
 * exportName or eqmName, but not both are optional input parameters.
 * \param fec (optional) is the returned FEC address structure of the server responding to the
 * call
 * \param srv (optional) is the returned Device Server Data address structure of the server responding
 * to the call.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \note A route for the services multicast group from the target server to the host using this call
 * must exist in order for this call to be successful.
 *
 */
TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecAddrStruct *fec,ExpDataStruct *srv);

/**
 * \fn TINE_EXPORT int GetCallerInformation(char *eqm,ClnInfo *clnInfoList,int *num)
 * \brief Returns the user name, network address and other information of all callers interested in the current contract
 *
 * \param eqm (in) is the equipment function name (local name) for which the exported name
 * is desired.
 * \param clnInfoList (out) if non-NULL is a reference to an array of ClnInfo objects to be filled with the caller list
 * \param num (in/out) is a reference to an int value which initially contains the size of the ClnInfoList array
 * used in the parameter list. Upon return of the call, it contains the actual number of callers.
 *
 * \note The clnInfoList supplied must point to buffer with enough room to hold at least the input 'num'
 * entries, otherwise a system crash is likely.
 *
 * \return 0 if successful, or a TINE completion code
 *
 * \see GetCaller()
 *
 * \b Example:
 *
 * \include eg_GetCallerInformation.c
 */
TINE_EXPORT int GetCallerInformation(char *eqm,ClnInfo *clnInfoList,int *num);

TINE_EXPORT void SetRunAsServer(void);
extern int gIsRunningAsServer;
extern int gCycleDeadbandFixed;

TINE_EXPORT void SetServerCycleIdle(int value);
TINE_EXPORT int GetServerCycleIdle(void);
TINE_EXPORT void SetClientCycleIdle(int value);
TINE_EXPORT int GetClientCycleIdle(void);

extern int gSubRenewalLength;
extern int gSubRenewalMultiplier;

/**
 * \brief Establishes whether a server can respond to multicast queries
 *  for its address.
 *
 * Normally a client acquires the address of a targeted server via
 * consulting the ENS (or its local address cache).  In some cases
 * it may not be possible to guarantee the availability of a running ENS.
 * If a server declares itself ready and capable of responding to a
 * 'where-are-you' multicast from a client application, it should call
 * this API routine and pass the value of 'TRUE'.  In this fashion it
 * is possible to have a client-server pair communicate with each other
 * in an 'ENS-less' scenario (provided the client and server exist on
 * a network which can forward the multicast probes issued by the client
 * (the multicast group 238.1.1.2).
 *
 * \param value is the desired setting
 * (default = FALSE)
 *
 * \see GetRespondToServiceRequests()
 *
 */
TINE_EXPORT void SetRespondToServiceRequests(int value);

/**
 * \brief Returns whether a server can respond to multicast queries
 *  for its address.
 *
 * Normally a client acquires the address of a targeted server via
 * consulting the ENS (or its local address cache).  In some cases
 * it may not be possible to guarantee the availability of a running ENS.
 * A server capable of responding to a
 * 'where-are-you' multicast from a client application will have
 * enabled this feature.
 *
 * \return the current setting
 * (default = FALSE)
 *
 * \see SetRespondToServiceRequests()
 *
 */
TINE_EXPORT int GetRespondToServiceRequests(void);

/**
 * \brief Sets the contract subscription renewal length
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.  The default value for subscription
 * renewals can be adjusted with this call.
 *
 * \param value is the desired contract renewal length.
 * (mininum = default = 60; maximum = 32767)
 *
 * \see GetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT void SetSystemSubscriptionRenewalLength(int value);

/**
 * \brief Gets the current contract subscription renewal length
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.  The default value for subscription
 * renewals can be retrieved with this call.
 *
 * \return the current setting of the contract renewal length
 *
 * \see SetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT int GetSystemSubscriptionRenewalLength(void);

/**
 * \brief Sets the current subscription renewal length for the property specified
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName is the registered property for the buffer is to be assigned.
 * \param value is the desired setting
 *
 * \return 0 upon success of a TINE error code
 *
 * \see SetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT int SetPropertySubscriptionRenewalLength(char *eqm,char *prpName,int value);

/**
 * \brief Gets the current subscription renewal length for the property specified
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.  The default value for subscription
 * renewals can be retrieved with this call.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName is the registered property for the buffer is to be assigned.
 * \param value is a reference to the integer variable to hold the returned setting.
 *
 * \return 0 upon success of a TINE error code
 *
 * \see GetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT int GetPropertySubscriptionRenewalLength(char *eqm,char *prpName,int *value);

/**
 * \brief Gets the current client-side subscription threshold for the link in question.
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.
 * The client side link can establish a higher renewal threshold value by making use of
 * this call.
 *
 * \param linkId is the link id for the link in question (returned from the original call to
 * AttachLink().
 * \param thresholdInPercent is the desired threshold given as a percent of the total number
 * of subscription transfers.  This should be positive integer.  Accepted values range
 * between 10 and 90 (percent).
 *
 * \return 0 upon success of a TINE error code
 *
 * \see GetSubscriptionRenewalThreshold()
 *
 */
TINE_EXPORT int SetSubscriptionRenewalThreshold(int linkId,int thresholdInPercent);

/**
 * \brief Gets the current client-side subscription threshold for the link in question.
 *
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the
 * subscription counter to descrease more rapidly than the client's
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.
 *
 * \param linkId is the link id for the link in question (returned from the original call to
 * AttachLink().
 * \param thesholdInPercent is a reference to hold the current setting.
 *
 * \return 0 upon success of a TINE error code
 *
 * \see SetSubscriptionRenewalThreshold()
 *
 */
TINE_EXPORT int GetSubscriptionRenewalThreshold(int linkId,int *thresholdInPercent);

/**
 * \fn TINE_EXPORT void SetPreSystemInitFunction(SYSTSKP fcn)
 * \brief Sets a user-specific initialization routine to be executed prior to
 * server initialization.
 *
 * \param fcn is the function to be called prior to system initialization.
 * (prototype void fcn(void)). Typically one places parameter settings which
 * differ from the default values in such a routine, so that they have
 * an effect on the server initialization.  Likewise, calls such as
 * RegisterFecInformation() should be made in such a routine.
 *
 * \see SetPostSystemInitFunction()
 *
 */
TINE_EXPORT void SetPreSystemInitFunction(SYSTSKP fcn);

/**
 * \fn TINE_EXPORT void SetPostSystemInitFunction(SYSTSKP fcn)
 * \brief Sets a user-specific initialization routine to be executed following
 * server initialization.
 *
 * \param fcn is the function to be called following system initialization.
 * (prototype void fcn(void)). Typically one places registration calls
 * such as RegisterEquipmentModule(), etc. in such a routine.
 *
 * \see SetPreSystemInitFunction()
 *
 */
TINE_EXPORT void SetPostSystemInitFunction(SYSTSKP fcn);

/**
 * \fn TINE_EXPORT void SetSystemCleanupFunction(SYSTSKP fcn)
 * \brief Sets a user-specific cleanup routine to be executed as a
 * final step during a normal cleanup phase (e.g. typing 'quit' or 'exit'
 * at the command prompt.
 *
 * \param fcn is the cleanup function to be called
 * (prototype void fcn(void)).
 *
 */
TINE_EXPORT void SetSystemCleanupFunction(SYSTSKP fcn);

/**
 * \fn TINE_EXPORT void SetDefaultTransportMode(int value)
 * \brief Sets the default TINE transport mode used in client-side links.
 *
 * \param value is the desired default transport mode.  This can be either
 * 'TCP' or 'UDP' (the default).  Indeed, if value does not equal 'TCP' then
 * the default transport mode will be taken as UDP.
 *
 * \see GetDefaultTransportMode()
 */
TINE_EXPORT void SetDefaultTransportMode(int value);

/**
 * \fn TINE_EXPORT int GetDefaultTransportMode(void)
 * \brief Gets the default TINE transport mode used in client-side links.
 *
 * \return the default TINE transport mode (either 'TCP' or 'UDP').
 *
 * \see SetDefaultTransportMode()
 */
TINE_EXPORT int GetDefaultTransportMode(void);

/**
 * \fn TINE_EXPORT int SetSystemAlias(char *alias, char *name)
 * \brief Sets an alias for either a registered property or registered device
 *
 * A client can make use of an alias in place of either a registered device name
 * or a registered property name in any call.  The alias name will be replaced
 * by its target within the contract tables.  Property aliases are also returned
 * in queries to obtain the registered property list.
 *
 * \param alias is the desired alias for the designated property or device
 * \param name is the name of the alias target (either a registered device
 * or a registered property).
 *
 * \return 0 upon success or a TINE error code.
 *
 */
TINE_EXPORT int SetSystemAlias(char *alias, char *name);

/**
 * \fn TINE_EXPORT char *GetSystemAlias(char *name)
 * \brief Gets the alias for either a registered property or registered device
 *
 * A client can make use of an alias in place of either a registered device name
 * or a registered property name in any call.  The alias name will be replaced
 * by its target within the contract tables.  Property aliases are also returned
 * in queries to obtain the registered property list.
 *
 * \param name is the name of the alias target (either a registered device
 * or a registered property).
 *
 * \return the registered alias for the given input or NULL if there is no registered alias.
 *
 */
TINE_EXPORT char *GetSystemAlias(char *name);

/**
 * \fn TINE_EXPORT char *ResolveSystemAlias(char *alias)
 * \brief Gets the registered property or registered device name for the given alias
 *
 * A client can make use of an alias in place of either a registered device name
 * or a registered property name in any call.  The alias name will be replaced
 * by its target within the contract tables.  Property aliases are also returned
 * in queries to obtain the registered property list.
 *
 * \param alias is the alias target for which the real registered name is desired
 * (either a registered device or a registered property).
 *
 * \return the registered name for the given input or NULL if the input is not a registered alias.
 *
 */
TINE_EXPORT char *ResolveSystemAlias(char *alias);

#define getMultiplierFromRate(r) ( (r) > 0 && (r) < 1000 ? 1000/(r) : 1)
#define getThisSubscriptionRenewalLength(r) (\
  gSubRenewalLength != CTR_RENEWAL ? gSubRenewalLength : CTR_RENEWAL * getMultiplierFromRate(r)\
)
#define getEffectiveRenewalMultiplier(m,r) (\
  (int)(m) > 1 ? (int)(m) : (int)getMultiplierFromRate(r)\
)

extern int gSystemDataStamp;
extern int gSystemDataStampRb;
extern int gSystemDataStampDly;
extern struct timeval gSystemDataStampTs;
void applySystemDataStamp(void);

extern char gDoocsUserName[USERNAME_SIZE+8];
TINE_EXPORT void SetDoocsUserName(const char *usr);
TINE_EXPORT char *GetDoocsUserName(void);
TINE_EXPORT void SetUserType(const char *type);
TINE_EXPORT char *GetUserType(void);
TINE_EXPORT void SetGlobalsType(const char *type);
TINE_EXPORT char *GetGlobalsType(void);
TINE_EXPORT void SetEnforceMultiChannelAccess(int value);
TINE_EXPORT int GetEnforceMultiChannelAccess(void);

TINE_EXPORT FecAddrStruct *GetFecAddrFromCache(char *context,char *expName);
TINE_EXPORT FecDataStruct *GetFecDataFromCache(char *context,char *expName);
TINE_EXPORT int GetFECAddr(char *fecName,FecDataStruct *fecaddr);
TINE_EXPORT int GetFecAddr(char *fecName,FecAddrStruct *fecaddr);
TINE_EXPORT int GetServerAddr(char *context,char *expName,ExpDataStruct *srvaddr,FecAddrStruct *fecaddr);
TINE_EXPORT int GetConnectionAddr(char *context,char *expName,char *devName,char *prpName);
TINE_EXPORT void CloseClientsEx(int closesockets,int quickExit);
TINE_EXPORT void CloseClients(int closesockets);
TINE_EXPORT void CloseClientLinks(void);
#define closeClients(a1) CloseClients(a1)
TINE_EXPORT void CloseServers(int closesockets);
#define closeServers(a1) CloseServers(a1)
TINE_EXPORT void UseMultiThreadedCycler(int value);
TINE_EXPORT void UseSynchronousNotification(int value);

/**
 * \brief sets the number of allocated buffers to use in contract access of a
 * a registered property.
 *
 * \param value is the number of buffers to use.  This may only be set to 1 or 2.
 * '1' is the default for single-threaded builds. For multi-threaded builds this
 * value is fixed to 2.
 *
 * \return the number of allocated buffers used.
 */
TINE_EXPORT int SetNumberDataOutputBuffers(int value);
/**
 * \brief gets the number of allocated buffers to use in contract access of a
 * a registered property.
 *
 * \return the number of allocated buffers used.
 */
TINE_EXPORT int GetNumberDataOutputBuffers(void);
/**
 * \brief returns a pointer to a useable reference pointer associated with
 * the contract currently being accessed.
 *
 * If some persistence is desired over the lifetime of a registered contract,
 * this function may be used to associate a specific reference.  A contract
 * will provide a reference pointer initialized to a NULL pointer.
 * By obtaining the pointer to this reference, a specific pointer can then be
 * assigned.
 * This is expected to be entirely managed by the user, who is then responsible
 * for all cleanup activity when the contract goes out of scope.
 *
 * \param el is a reference to the targeted Export List Structure
 * (obtained via a call to GetExportListItem()).
 *
 * \return a pointer to the specific reference of the currently executing contract, or
 * NULL if the contract is NOT being currently executed or the given equipment module name
 * does not refer to a registered equipment module.
 *
 * \see GetContractDataReferenceByEqmName
 */
TINE_EXPORT void **GetContractDataReference(ExportListStruct *el);

/**
 * \brief returns a pointer to a useable reference pointer associated with
 * the contract currently being accessed.
 *
 * If some persistence is desired over the lifetime of a registered contract,
 * this function may be used to associate a specific reference.  A contract
 * will provide a reference pointer initialized to a NULL pointer.
 * By obtaining the pointer to this reference, a specific pointer can then be
 * assigned.
 * This is expected to be entirely managed by the user, who is then responsible
 * for all cleanup activity when the contract goes out of scope.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return a pointer to the specific reference of the currently executing contract, or
 * NULL if the contract is NOT being currently executed or the given equipment module name
 * does not refer to a registered equipment module.
 *
 * \see GetContractDataReference
 */
TINE_EXPORT void **GetContractDataReferenceByEqmName(char *eqm);

int getClockTickMilliseconds(void);
/**
 * \fn TINE_EXPORT int microsleep(int usecs)
 * \brief sleep for given number of micro-seconds
 *
 * If the input is less than 4 micro-seconds the delay
 * is a 'hard-spin' where the OS system clock is examined in
 * a loop without yielding the time slice.  If the input is greater
 * than 4 but less than 50 micro-seconds the delay is a 'soft-spin'
 * where the OS system clock is examined in a loop but yielding
 * the thread's time slice at each pass through the loop.
 * If the input is greater than 50 micro-seconds the delay
 * invokes a nanosleep or equivalent which yields the thread
 * until the delay time expires.  Unless a real-time clock
 * is used (e.g. VxWorks) the granularity of this latter
 * technique is never better than 50 micro-seconds (often
 * more like a clock tick).
 *
 * \param usecs the desired number of micro-seconds to sleep
 *
 * \return the number of micro seconds slept
 */
TINE_EXPORT int microsleep(int usecs);
extern int heapCorrupted;
#if defined(WIN32) && !defined(_DEBUG)
# define assertHeapOkay(t,d,p) {\
  if (_heapchk() != _HEAPOK && !heapCorrupted) {\
    feclogEx(TINE_LOGLEVEL_ERROR, "%s %s %s : memory overwrite has occurred",t,d,p);\
    msglog(memory_overwrite,"%s %s %s: memory overwrite has occurred",t,d,p);\
    heapCorrupted = TRUE;\
  }\
}
#else
# define assertHeapOkay(t,d,p) ;
#endif
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif  /* SRVCORE_H_ */
